# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.cli
    :platform: Unix
    :synopsis: Burp-UI CLI module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys
import time
import click

if os.getenv("BUI_MODE") in ["server", "ws"] or "websocket" in sys.argv:
    try:
        from gevent import monkey

        monkey.patch_socket()
    except ImportError:
        pass

from .app import create_app  # noqa
from .exceptions import BUIserverException  # noqa

try:
    from flask_socketio import SocketIO  # noqa

    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

ROOT = os.path.dirname(os.path.realpath(__file__))
DEBUG = os.getenv("BUI_DEBUG") or os.getenv("FLASK_DEBUG") or False
DEBUG = DEBUG and DEBUG.lower() not in ["false", "no", "0"]

VERBOSE = os.getenv("BUI_VERBOSE") or 0
if VERBOSE:
    try:
        VERBOSE = int(VERBOSE)
    except ValueError:
        VERBOSE = 0

# UNITTEST is used to skip the burp-2 requirements for modes != server
UNITTEST = os.getenv("BUI_MODE") not in ["server", "manage", "celery", "legacy", "ws"]
CLI = os.getenv("BUI_MODE") not in ["server", "legacy"]

app = create_app(
    conf=os.environ.get("BUI_CONFIG"),
    verbose=VERBOSE,
    logfile=os.environ.get("BUI_LOGFILE"),
    debug=DEBUG,
    gunicorn=False,
    unittest=UNITTEST,
    cli=CLI,
    websocket_server=(os.getenv("BUI_MODE") == "ws" or "websocket" in sys.argv),
)

try:
    from .extensions import create_db
    from .ext.sql import db
    from flask_migrate import Migrate

    # This may have been reseted by create_app
    if isinstance(app.database, bool):
        app.config["WITH_SQL"] = app.database
    else:
        app.config["WITH_SQL"] = app.database and app.database.lower() != "none"

    if app.config["WITH_SQL"]:
        create_db(app, True)

        mig_dir = os.getenv("BUI_MIGRATIONS")
        if mig_dir:
            migrate = Migrate(app, db, mig_dir)
        else:
            migrate = Migrate(app, db)
except ImportError:
    pass

# Utilities functions


def _die(error, appli=None):
    appli = " '{}'".format(appli) if appli else ""
    err(
        "Unable to initialize the application{}: {}".format(appli, error),
    )
    sys.exit(2)


def _log(message, *args, color=None, err=False):
    msg = message % args if args else message
    if color is not None:
        msg = click.style(msg, fg=color)
    click.echo(msg, err=err)


def log(message="", *args):
    _log(message, *args)


def ok(message="", *args):
    _log(message, *args, color="green")


def info(message="", *args):
    _log(message, *args, color="blue")


def warn(message="", *args):
    _log(message, *args, color="yellow")


def err(message="", *args, err=True):
    _log(message, *args, color="red", err=err)


@app.cli.command()
def legacy():
    """Legacy server for backward compatibility."""
    warn(
        "If you want to pass options, you should run 'python -m burpui "
        "-m legacy [...]' instead",
    )
    app.manual_run()


@app.cli.command()
@click.option(
    "-b",
    "--bind",
    default="127.0.0.1",
    help="Which address to bind to for the websocket server",
)
@click.option(
    "-p",
    "--port",
    default=5001,
    help="Which port to listen on for the websocket server",
)
@click.option(
    "-d",
    "--debug",
    default=False,
    is_flag=True,
    help="Whether to start the websocket server in debug mode",
)
def websocket(bind, port, debug):
    """Start a new websocket server."""
    try:
        from .ext.ws import socketio
    except ImportError:
        _die(
            "Missing requirement, did you ran 'pip install" ' "burp-ui[websocket]"\'?',
            "websocket",
        )
    socketio.run(app, host=bind, port=port, debug=debug)


@app.cli.command()
@click.option(
    "-b", "--backend", default="BASIC", help="User Backend (default is BASIC)."
)
@click.option("-p", "--password", help="Password to assign to user.", default=None)
@click.option(
    "-a",
    "--ask",
    default=False,
    is_flag=True,
    help="If no password is provided and this flag is enabled, "
    "you'll be prompted for one, else a random one will be "
    "generated.",
)
@click.option(
    "-v", "--verbose", default=False, is_flag=True, help="Add extra debug messages."
)
@click.argument("name")
def create_user(backend, password, ask, verbose, name):
    """Create a new user."""
    try:
        msg = app.load_modules()
    except Exception as e:
        msg = str(e)

    if not backend.endswith(":AUTH"):
        backend = f"{backend}:AUTH"
    backend = backend.upper()

    if msg:
        _die(msg, "create_user")

    info("[*] Adding '{}' user...".format(name))
    try:
        handler = getattr(app, "uhandler")
    except AttributeError:
        handler = None

    if not handler or len(handler.backends) == 0 or backend not in handler.backends:
        err("[!] No authentication backend found")
        sys.exit(1)

    back = handler.backends[backend]

    if back.add_user is False:
        err("[!] The '{}' backend does not support user creation".format(backend))
        sys.exit(2)

    if not password:
        if ask:
            import getpass

            password = getpass.getpass()
            confirm = getpass.getpass("Confirm: ")
            if password != confirm:
                err("[!] Passwords mismatch")
                sys.exit(3)
        else:
            import random

            alphabet = (
                "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLM" "NOPQRSTUVWXYZ"
            )
            pw_length = 8
            mypw = ""

            for i in range(pw_length):
                next_index = random.randrange(len(alphabet))
                mypw += alphabet[next_index]
            password = mypw
            info("[+] Generated password: {}".format(password))

    success, message, _ = back.add_user(name, password)
    _log(
        "[+] Success: {}{}".format(
            success, " -> {}".format(message) if verbose and message else ""
        ),
        color="green" if success else "red",
    )


@app.cli.command()
@click.option("-p", "--password", help="Password to assign to user.", default=None)
@click.option(
    "-u",
    "--username",
    help="Provide the username to get the full " "configuration line.",
    default=None,
)
@click.option(
    "-b",
    "--batch",
    default=False,
    is_flag=True,
    help="Don't be extra verbose so that you can use the output "
    "directly in your scripts. Requires both -u and -p.",
)
def hash_password(password, username, batch):
    """Hash a given password to fill the configuration file."""
    from werkzeug.security import generate_password_hash

    if batch and (not username or not password):
        err(
            "You need to provide both a username and a password using the "
            "-u and -p flags!",
        )
        sys.exit(1)

    askpass = False
    if not password:
        askpass = True
        import getpass

        password = getpass.getpass()

    hashed = generate_password_hash(password)
    if not batch:
        log("'{}' hashed into: {}".format(password if not askpass else "*" * 8, hashed))
    if username:
        if not batch:
            info("#8<{}".format("-" * 77))
        log("{} = {}".format(username, hashed))
        if not batch:
            info("#8<{}".format("-" * 77))


@app.cli.command()
@click.argument("language")
def init_translation(language):
    """Initialize a new translation for the given language."""
    try:
        import babel  # noqa
    except ImportError:
        warn("Missing i18n requirements, giving up")
        return
    os.chdir(os.path.join(ROOT, ".."))
    os.system(
        "pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui"
    )
    os.system(
        "pybabel init -i messages.pot -d burpui/translations -l {}".format(language)
    )
    os.unlink("messages.pot")


@app.cli.command()
def update_translation():
    """Update translation files."""
    try:
        import babel  # noqa
    except ImportError:
        warn("Missing i18n requirements, giving up")
        return
    os.chdir(os.path.join(ROOT, ".."))
    os.system(
        "pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui"
    )
    os.system("pybabel update -i messages.pot -d burpui/translations")
    os.unlink("messages.pot")


@app.cli.command()
def compile_translation():
    """Compile translations."""
    try:
        import babel  # noqa
    except ImportError:
        warn("Missing i18n requirements, giving up")
        return
    os.chdir(os.path.join(ROOT, ".."))
    os.system("pybabel compile -f -d burpui/translations")


@app.cli.command()
@click.option(
    "-b",
    "--burp-conf-cli",
    "bconfcli",
    default=None,
    help="Burp client configuration file",
)
@click.option(
    "-s",
    "--burp-conf-serv",
    "bconfsrv",
    default=None,
    help="Burp server configuration file",
)
@click.option(
    "-c",
    "--client",
    default="bui",
    help="Name of the burp client that will be used by Burp-UI " '(defaults to "bui")',
)
@click.option(
    "-l",
    "--listen",
    default="0.0.0.0:5971",
    help="Setup a custom listen port for the Burp-UI restorations",
)
@click.option(
    "-h",
    "--host",
    default="::1",
    help='Address of the status server (defaults to "::1")',
)
@click.option("-r", "--redis", default=None, help="Redis URL to connect to")
@click.option(
    "-d",
    "--database",
    default=None,
    help="Database to connect to for persistent storage",
)
@click.option("-p", "--plugins", default=None, help="Plugins location")
@click.option("-m", "--monitor", default=None, help="bui-monitor configuration file")
@click.option(
    "-i", "--monitor-listen", "mbind", default=None, help="bui-monitor bind address"
)
@click.option(
    "-C",
    "--concurrency",
    default=None,
    type=click.INT,
    help="Number of concurrent requests addressed to the monitor",
)
@click.option(
    "-P",
    "--pool-size",
    "pool",
    default=None,
    type=click.INT,
    help="Number of burp-client processes to spawn in the monitor",
)
@click.option(
    "-B",
    "--backend",
    default=None,
    help="Switch to another backend",
    type=click.Choice(["burp2", "parallel"]),
)
@click.option(
    "-a",
    "--assume-version",
    "assume",
    default=None,
    help="If we cannot determine server version use this one",
)
@click.option(
    "-n",
    "--dry",
    is_flag=True,
    help="Dry mode. Do not edit the files but display changes",
)
def setup_burp(
    bconfcli,
    bconfsrv,
    client,
    listen,
    host,
    redis,
    database,
    plugins,
    monitor,
    mbind,
    concurrency,
    pool,
    backend,
    assume,
    dry,
):
    """Setup burp client for burp-ui."""
    if app.config["BACKEND"] not in ["burp2", "parallel"] and not backend:
        err("Sorry, you can only setup the 'burp2' and the 'parallel' backends")
        sys.exit(1)

    if not app.config["STANDALONE"]:
        err("Sorry, only the standalone mode is supported")
        sys.exit(1)

    if concurrency:
        from multiprocessing import cpu_count

        if concurrency > cpu_count():
            warn(
                "Warning: setting a concurrency level higher than the available CPU"
                " count might cause you some troubles"
            )

    if pool and concurrency:
        if concurrency > pool:
            warn(
                "Warning: setting a concurrency level higher than the available"
                " processes in the monitor is not recommended"
            )

    is_parallel = app.config["BACKEND"] == "parallel" or (
        backend and backend == "parallel"
    )

    # enforce burp2 backend for the configuration
    app.config["BACKEND"] = "burp2"

    try:
        msg = app.load_modules()
    except Exception as e:
        msg = str(e)

    if msg:
        _die(msg, "setup-burp")

    from .misc.parser.utils import Config
    from .misc.backend.utils.constant import BURP_LISTEN_OPTION, BURP_BIND_MULTIPLE
    from .app import get_redis_server
    from .config import BUIConfig
    import difflib
    import tempfile

    if monitor:
        monconf = BUIConfig(monitor)
        monconf_orig = []
        mon_orig = mon_source = monitor

        if dry:
            try:
                with open(monitor) as fil:
                    monconf_orig = fil.readlines()
            except:
                pass

            (_, temp) = tempfile.mkstemp()
            monconf.options.filename = temp

    parser = app.client.get_parser()
    server_version = app.client.get_server_version() or assume
    orig = source = None
    conf_orig = []
    if dry:
        try:
            with open(app.conf.options.filename) as fil:
                conf_orig = fil.readlines()
        except:
            pass

        orig = source = app.conf.options.filename
        (_, temp) = tempfile.mkstemp()
        app.conf.options.filename = temp

    # handle migration of old config files
    if app.conf.section_exists("Burp2"):
        if app.conf.rename_section("Burp2", "Burp", source):
            info("Renaming old [Burp2] section")
            app.conf._refresh(True)

    refresh = False
    if not app.conf.lookup_section("Burp", source):
        refresh = True
    if not app.conf.lookup_section("Global", source):
        refresh = True
    if (database or redis) and not app.conf.lookup_section("Production", source):
        refresh = True
    if concurrency and not app.conf.lookup_section("Parallel", source):
        refresh = True

    if refresh:
        app.conf._refresh(True)

    if monitor and not monconf.lookup_section("Global", mon_source):
        monconf._refresh(True)

    def _edit_conf(key, val, attr, section="Burp", obj=app.client, conf=app.conf):
        if val and (
            (key not in conf.options[section])
            or (key in conf.options[section] and val != conf.options[section][key])
        ):
            adding = key not in conf.options[section]
            conf.options[section][key] = val
            conf.options.write()
            if obj:
                setattr(obj, attr, val)
            if adding:
                msg = f'Adding new option "{key}={val}" to section [{section}] in {conf.conffile}'
            else:
                msg = f'Updating option "{key}={val}" in section [{section}] in {conf.conffile}'
            info(msg)
            return True
        return False

    def _color_diff(line):
        if line.startswith("+"):
            return click.style(line, fg="green")
        elif line.startswith("-"):
            return click.style(line, fg="red")
        elif line.startswith("^"):
            return click.style(line, fg="blue")
        return line

    refresh = False
    refresh |= _edit_conf("bconfcli", bconfcli, "burpconfcli")
    refresh |= _edit_conf("bconfsrv", bconfsrv, "burpconfsrv")
    refresh |= _edit_conf("plugins", plugins, "plugins", "Global", app)
    if backend:
        refresh |= _edit_conf("backend", backend, None, "Global", None)
    if is_parallel and concurrency:
        refresh |= _edit_conf("concurrency", concurrency, None, "Parallel", None)
    if mbind:
        refresh |= _edit_conf("host", mbind, None, "Parallel", None)

    if refresh:
        app.conf._refresh(True)

    refresh = False
    if monitor and pool:
        refresh |= _edit_conf("pool", pool, None, "Global", None, monconf)
    if monitor and mbind:
        refresh |= _edit_conf("bind", mbind, None, "Global", None, monconf)
    if monitor:
        refresh |= _edit_conf("bconfcli", bconfcli, None, "Burp", None, monconf)

    if refresh:
        monconf._refresh(True)

    if monitor and app.config["BACKEND"] == "parallel":
        mon_password = monconf.options["Global"].get("password")
        back_password = app.conf.options["Parallel"].get("password")

        if mon_password != back_password:
            click.echo(
                click.style(
                    "Backend password does not match monitor password", fg="yellow"
                )
            )

    if redis:
        try:
            # detect missing modules
            import redis as redis_client  # noqa
            import celery  # noqa
            import socket

            if (
                "redis" not in app.conf.options["Production"]
                or "redis" in app.conf.options["Production"]
                and app.conf.options["Production"]["redis"] != redis
            ) and app.redis != redis:
                app.conf.options["Production"]["redis"] = redis
                app.redis = redis

            rhost, rport, _ = get_redis_server(app)
            ret = -1
            for _ in range(10):
                if ret == 0:
                    break

                for res in socket.getaddrinfo(
                    rhost, rport, socket.AF_UNSPEC, socket.SOCK_STREAM
                ):
                    if ret == 0:
                        break
                    af, socktype, proto, _, sa = res
                    try:
                        s = socket.socket(af, socktype, proto)
                    except socket.error:
                        continue
                    try:
                        ret = s.connect_ex(sa)
                    except:
                        continue

                time.sleep(1)

            if ret == 0:
                app.conf.options["Production"]["celery"] = "true"

                app.conf.options["Production"]["storage"] = "redis"

                app.conf.options["Production"]["cache"] = "redis"
            else:
                warn("Unable to contact the redis server, disabling it")
                app.conf.options["Production"]["storage"] = "default"
                app.conf.options["Production"]["cache"] = "default"
                if app.use_celery:
                    app.conf.options["Production"]["celery"] = "false"

            app.conf.options.write()
            app.conf._refresh(True)
        except ImportError:
            warn(
                "Unable to activate redis & celery. Did you ran the "
                "'pip install burp-ui[celery]' and "
                "'pip install burp-ui[gunicorn-extra]' commands first?"
            )

    if database:
        try:
            from .ext.sql import db  # noqa

            if (
                "database" not in app.conf.options["Production"]
                or "database" in app.conf.options["Production"]
                and app.conf.options["Production"]["database"] != database
            ) and app.database != database:
                app.conf.options["Production"]["database"] = database
                app.conf.options.write()
                app.conf._refresh(True)
        except ImportError:
            warn(
                "It looks like some dependencies are missing. Did you ran "
                "the 'pip install \"burp-ui[sql]\"' command first?"
            )

    if dry:
        temp = app.conf.options.filename
        app.conf.options.filename = orig
        after = []
        try:
            if not os.path.exists(temp) or os.path.getsize(temp) == 0:
                after = conf_orig
            else:
                with open(temp) as fil:
                    after = fil.readlines()
                os.unlink(temp)
        except:
            pass
        diff = difflib.unified_diff(
            conf_orig, after, fromfile=orig, tofile="{}.new".format(orig)
        )
        out = ""
        for line in diff:
            out += _color_diff(line)
        if out:
            click.echo_via_pager(out)

    if dry and monitor:
        temp = monconf.options.filename
        monconf.options.filename = mon_orig
        after = []
        try:
            if not os.path.exists(temp) or os.path.getsize(temp) == 0:
                after = monconf_orig
            else:
                with open(temp) as fil:
                    after = fil.readlines()
                os.unlink(temp)
        except:
            pass
        diff = difflib.unified_diff(
            monconf_orig, after, fromfile=mon_orig, tofile="{}.new".format(mon_orig)
        )
        out = ""
        for line in diff:
            out += _color_diff(line)
        if out:
            click.echo_via_pager(out)

    bconfcli = (
        bconfcli
        or app.conf.options["Burp"].get("bconfcli")
        or getattr(app.client, "burpconfcli")
    )
    bconfsrv = (
        bconfsrv
        or app.conf.options["Burp"].get("bconfsrv")
        or getattr(app.client, "burpconfsrv")
    )
    dest_bconfcli = bconfcli

    is_burp_2_2_10_plus = False
    listen_opt = "status_address"
    if server_version and server_version >= BURP_LISTEN_OPTION:
        is_burp_2_2_10_plus = True
        listen_opt = "listen_status"

    _, restore_port = listen.split(":")

    if not os.path.exists(bconfcli):
        clitpl = f"""
mode = client
port = {restore_port}
status_port = 4972
server = ::1
password = abcdefgh
cname = {client}
protocol = 1
pidfile = /tmp/burp.client.pid
syslog = 0
stdout = 1
progress_counter = 1
network_timeout = 72000
server_can_restore = 0
cross_all_filesystems=0
ca_burp_ca = /usr/sbin/burp_ca
ca_csr_dir = /etc/burp/CA-client
ssl_cert_ca = /etc/burp/ssl_cert_ca-client-{client}.pem
ssl_cert = /etc/burp/ssl_cert-bui-client.pem
ssl_key = /etc/burp/ssl_cert-bui-client.key
ssl_key_password = password
ssl_peer_cn = burpserver
include = /home
exclude_fs = sysfs
exclude_fs = tmpfs
nobackup = .nobackup
exclude_comp=bz2
exclude_comp=gz
"""

        if dry:
            (_, dest_bconfcli) = tempfile.mkstemp()
        with open(dest_bconfcli, "w") as confcli:
            confcli.write(clitpl)

    parser = app.client.get_parser(assume)

    confcli = Config(dest_bconfcli, parser, "srv")
    confcli.set_default(dest_bconfcli)
    confcli.parse()

    if confcli.get("cname") != client:
        confcli["cname"] = client
    if confcli.get("server") != host:
        confcli["server"] = host
    c_status_port = (
        confcli.get("status_port", [4972])[0]
        if confcli.version >= BURP_BIND_MULTIPLE
        else confcli.get("status_port", 4972)
    )
    c_server_port = (
        confcli.get("port", [4971])[0]
        if confcli.version >= BURP_BIND_MULTIPLE
        else confcli.get("port", 4971)
    )
    if c_server_port != restore_port:
        confcli["port"] = [restore_port]

    if confcli.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfcli

        confcli.store(conf=bconfcli, dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfcli) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass

            if dest_bconfcli != bconfcli:
                # the file did not exist
                os.unlink(dest_bconfcli)
                before = []

            diff = difflib.unified_diff(
                before, after, fromfile=bconfcli, tofile="{}.new".format(bconfcli)
            )
            out = ""
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if not os.path.exists(bconfsrv):
        err("Unable to locate burp-server configuration, aborting!")
        sys.exit(1)

    confsrv = Config(bconfsrv, parser, "srv")
    confsrv.set_default(bconfsrv)
    confsrv.parse()

    bind_index = -1

    if host not in ["::1", "127.0.0.1"]:
        bind = confsrv.get(listen_opt)
        if is_burp_2_2_10_plus:
            bind_list = list(bind)
            for idx, line in enumerate(bind_list):
                if line.endswith(":{}".format(c_status_port)):
                    bind = line.split(":")[0]
                    bind_index = idx
                    break
            else:
                warn(
                    "Unable to locate a 'listen_status' associated to your client "
                    "'status_port' ({})".format(c_status_port)
                )
        if (bind and bind not in [host, "::", "0.0.0.0"]) or not bind:
            warn(
                "It looks like your burp server is not exposing it's "
                "status port in a way that is reachable by Burp-UI!"
            )
            info(
                "You may want to set the '{0}' setting with "
                "either '{1}{3}', '::{3}' or '0.0.0.0{3}' in the {2} file "
                "in order to make Burp-UI work".format(
                    listen_opt,
                    host,
                    bconfsrv,
                    ":{}".format(c_status_port) if is_burp_2_2_10_plus else "",
                )
            )

    MAX_STATUS_CHILDREN = pool if pool is not None else 15
    if not is_burp_2_2_10_plus:
        s_port = confsrv.get("port", [4971])
        if restore_port not in s_port:
            confsrv["port"] = restore_port
    else:
        s_listen = confsrv.get("listen", [])
        if listen not in s_listen:
            confsrv["listen"] = listen
    status_port = confsrv.get("status_port", [4972])
    do_warn = False
    if "max_status_children" not in confsrv:
        info(
            "We need to set the number of 'max_status_children'. "
            "Setting it to {}.".format(MAX_STATUS_CHILDREN)
        )
        confsrv["max_status_children"] = MAX_STATUS_CHILDREN
    elif confsrv.version and confsrv.version < BURP_BIND_MULTIPLE:
        max_status_children = confsrv.get("max_status_children")
        if not max_status_children or max_status_children < MAX_STATUS_CHILDREN:
            confsrv["max_status_children"] = MAX_STATUS_CHILDREN
            do_warn = True
    else:
        max_status_children = confsrv.get("max_status_children", [])
        if not is_burp_2_2_10_plus:
            for idx, value in enumerate(status_port):
                if value == c_status_port:
                    bind_index = idx
                    break
        if bind_index >= 0 and len(max_status_children) >= bind_index:
            if max_status_children[bind_index] < MAX_STATUS_CHILDREN:
                confsrv["max_status_children"][bind_index] = MAX_STATUS_CHILDREN
                do_warn = True
        else:
            if max_status_children[-1] < MAX_STATUS_CHILDREN:
                confsrv["max_status_children"][-1] = MAX_STATUS_CHILDREN
                do_warn = True

    if do_warn:
        warn(
            "We need to raise the number of 'max_status_children' to {}.".format(
                MAX_STATUS_CHILDREN
            )
        )

    if "restore_client" not in confsrv:
        confsrv["restore_client"] = client
    else:
        restore = confsrv.getlist("restore_client")
        if client not in restore:
            confsrv["restore_client"].append(client)

    confsrv["monitor_browse_cache"] = True

    ca_client_dir = confsrv.get("ca_csr_dir")
    if ca_client_dir and not os.path.exists(ca_client_dir):
        try:
            os.makedirs(ca_client_dir)
        except IOError as exp:
            _log(
                'Unable to create "{}" dir: {}'.format(ca_client_dir, exp),
                color="yellow",
                err=True,
            )

    if confsrv.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfsrv

        confsrv.store(conf=bconfsrv, dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfsrv) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass
            diff = difflib.unified_diff(
                before, after, fromfile=bconfsrv, tofile="{}.new".format(bconfsrv)
            )
            out = ""
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if confsrv.get("clientconfdir"):
        bconfagent = os.path.join(confsrv.get("clientconfdir"), client)
    else:
        warn(
            'Unable to find "clientconfdir" option, you will have to '
            "setup the agent by your own"
        )
        bconfagent = os.devnull

    if not os.path.exists(bconfagent):

        agenttpl = """
password = abcdefgh
"""

        if not dry:
            with open(bconfagent, "w") as confagent:
                confagent.write(agenttpl)
        else:
            before = []
            after = ["{}\n".format(x) for x in agenttpl.splitlines()]
            diff = difflib.unified_diff(
                before, after, fromfile="None", tofile=bconfagent
            )
            out = ""
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    else:
        confagent = Config(bconfagent, parser, "cli")
        confagent.set_default(bconfagent)
        confagent.parse()

        if confagent.get("password") != confcli.get("password"):
            warn(
                "It looks like the passwords in the {} and the {} files "
                "mismatch. Burp-UI will not work properly until you fix "
                "this".format(bconfcli, bconfagent)
            )


@app.cli.command()
@click.option(
    "-c",
    "--client",
    default="bui",
    help="Name of the burp client that will be used by Burp-UI " '(defaults to "bui")',
)
@click.option(
    "-h",
    "--host",
    default="::1",
    help='Address of the status server (defaults to "::1")',
)
@click.option("-t", "--tips", is_flag=True, help="Show you some tips")
def diag(client, host, tips):
    """Check Burp-UI is correctly setup."""
    if app.config["BACKEND"] not in ["burp2", "parallel"]:
        err("Sorry, you can only diag the 'burp2' and the 'parallel' backends")
        sys.exit(1)

    if not app.config["STANDALONE"]:
        err("Sorry, only the standalone mode is supported")
        sys.exit(1)

    try:
        msg = app.load_modules()
    except Exception as e:
        msg = str(e)

    if msg:
        _die(msg, "diag")

    from .misc.backend.utils.constant import BURP_LISTEN_OPTION
    from .misc.parser.utils import Config
    from .app import get_redis_server

    def _value_in_option(value, option, section="Production"):
        if section not in app.conf.options:
            return False
        if option not in app.conf.options[section]:
            return False
        return value in app.conf.options[section][option]

    if (
        "Production" in app.conf.options
        and "redis" in app.conf.options["Production"]
        and any(_value_in_option("redis", x) for x in ["storage", "session", "cache"])
    ):
        try:
            # detect missing modules
            import redis as redis_client  # noqa
            import celery  # noqa
            import socket

            rhost, rport, _ = get_redis_server(app)
            ret = -1
            for res in socket.getaddrinfo(
                rhost, rport, socket.AF_UNSPEC, socket.SOCK_STREAM
            ):
                if ret == 0:
                    break
                af, socktype, proto, _, sa = res
                try:
                    s = socket.socket(af, socktype, proto)
                except socket.error:
                    continue
                try:
                    ret = s.connect_ex(sa)
                except:
                    continue

            if ret != 0:
                warn("Unable to contact the redis server, disabling it")

        except ImportError:
            warn(
                "Unable to activate redis & celery. Did you ran the "
                "'pip install \"burp-ui[celery]\"' and "
                "'pip install \"burp-ui[gunicorn-extra]\"' commands first?"
            )

    if (
        "Production" in app.conf.options
        and "database" in app.conf.options["Production"]
        and app.conf.options["Production"]["database"] != "none"
    ):
        try:
            from .ext.sql import db  # noqa
        except ImportError:
            warn(
                "It looks like some dependencies are missing. Did you ran "
                "the 'pip install \"burp-ui[sql]\"' command first?"
            )

    section = "Burp"
    if not app.conf.section_exists(section):
        warn("Section [Burp] not found, looking for the old [Burp2] section instead.")
        section = "Burp2"
        if not app.conf.section_exists(section):
            err("No [Burp*] section found at all!", err=False)
            section = "Burp"

    bconfcli = app.conf.options.get(section, {}).get("bconfcli") or getattr(
        app.client, "burpconfcli"
    )
    bconfsrv = app.conf.options.get(section, {}).get("bconfsrv") or getattr(
        app.client, "burpconfsrv"
    )

    try:
        app.client.status()
    except Exception as e:
        if "Unable to spawn burp process" in str(e):
            try:
                app.client._spawn_burp(verbose=True)
            except Exception as e:
                msg = str(e)
        else:
            msg = str(e)

    if msg:
        err(msg, err=False)
        if "could not connect" in msg:
            warn(
                "It looks like your burp-client can not reach your "
                "burp-server. Please check both your 'server' setting in "
                "your '{}' file and 'status_address' in your '{}' "
                "file.".format(bconfcli, bconfsrv)
            )

    c_status_port = 4972
    errors = False
    if os.path.exists(bconfcli):
        try:
            parser = app.client.get_parser()

            confcli = Config(bconfcli, parser, "srv")
            confcli.set_default(bconfcli)
            confcli.parse()

            c_status_port = confcli.get("status_port", [4972])[0]

            if confcli.get("cname") != client:
                warn(
                    "The cname of your burp client does not match: "
                    "{} != {}".format(confcli.get("cname"), client)
                )
                errors = True
            if confcli.get("server") != host:
                warn(
                    "The burp server address does not match: "
                    "{} != {}".format(confcli.get("server"), host)
                )
                errors = True
        except BUIserverException as exc:
            err(str(exc))
            errors = True

    else:
        err("No client conf file found: {} does not exist".format(bconfcli))
        errors = True

    if os.path.exists(bconfsrv):
        try:
            is_burp_2_2_10_plus = False
            listen_opt = "status_address"
            server_version = app.client.get_server_version()
            if server_version and server_version >= BURP_LISTEN_OPTION:
                is_burp_2_2_10_plus = True
                listen_opt = "listen_status"

            parser = app.client.get_parser()

            confsrv = Config(bconfsrv, parser, "srv")
            confsrv.set_default(bconfsrv)
            confsrv.parse()

            bind_index = -1

            if host not in ["::1", "127.0.0.1"]:
                bind = confsrv.get(listen_opt)
                if is_burp_2_2_10_plus:
                    bind_list = list(bind)
                    for idx, line in enumerate(bind_list):
                        if line.endswith(":{}".format(c_status_port)):
                            bind = line.split(":")[0]
                            bind_index = idx
                            break
                    else:
                        warn(
                            "Unable to locate a 'listen_status' associated to your client "
                            "'status_port' ({})".format(c_status_port)
                        )
                if (bind and bind not in [host, "::", "0.0.0.0"]) or not bind:
                    warn(
                        "It looks like your burp server is not exposing it's "
                        "status port in a way that is reachable by Burp-UI!"
                    )
                    info(
                        "You may want to set the '{0}' setting with "
                        "either '{1}{3}', '::{3}' or '0.0.0.0{3}' in the {2} file "
                        "in order to make Burp-UI work".format(
                            listen_opt,
                            host,
                            bconfsrv,
                            ":{}".format(c_status_port) if is_burp_2_2_10_plus else "",
                        )
                    )
                    errors = True

            status_port = confsrv.get("status_port", [4972])
            if "max_status_children" not in confsrv:
                do_warn = True
            else:
                max_status_children = confsrv.get("max_status_children", [])
                do_warn = False
                if not is_burp_2_2_10_plus:
                    for idx, value in enumerate(status_port):
                        if value == c_status_port:
                            bind_index = idx
                            break
                if bind_index >= 0 and len(max_status_children) >= bind_index:
                    if max_status_children[bind_index] < 15:
                        do_warn = True
                else:
                    if max_status_children[-1] < 15:
                        do_warn = True
            if do_warn:
                info(
                    "'max_status_children' is to low, you need to set it to "
                    "15 or more. Please edit your {} file.".format(bconfsrv)
                )
                errors = True

            restore = []
            if "restore_client" in confsrv:
                restore = confsrv.getlist("restore_client")

            if client not in restore:
                warn(
                    "Your burp client is not listed as a 'restore_client'. "
                    "You won't be able to view other clients stats!"
                )
                errors = True

            if "monitor_browse_cache" not in confsrv or not confsrv.get(
                "monitor_browse_cache"
            ):
                warn(
                    "For performance reasons, it is recommended to enable the "
                    "'monitor_browse_cache'."
                )
                errors = True

            ca_client_dir = confsrv.get("ca_csr_dir")
            if ca_client_dir and not os.path.exists(ca_client_dir):
                try:
                    os.makedirs(ca_client_dir)
                except IOError as exp:
                    _log(
                        'Unable to create "{}" dir: {}'.format(ca_client_dir, exp),
                        color="yellow",
                        err=True,
                    )

            if confsrv.get("clientconfdir"):
                bconfagent = os.path.join(confsrv.get("clientconfdir"), client)
            else:
                warn(
                    'Unable to find "clientconfdir" option. Something is wrong '
                    "with your setup."
                )
                bconfagent = "ihopethisfiledoesnotexistbecauseitisrelatedtoburpui"

            if not os.path.exists(bconfagent) and bconfagent.startswith("/"):
                warn("Unable to find the {} file.".format(bconfagent))
                errors = True
            else:
                confagent = Config(bconfagent, parser, "cli")
                confagent.set_default(bconfagent)
                confagent.parse()

                if confagent.get("password") != confcli.get("password"):
                    warn(
                        "It looks like the passwords in the {} and the {} files "
                        "mismatch. Burp-UI will not work properly until you fix "
                        "this.".format(bconfcli, bconfagent)
                    )
        except BUIserverException as exc:
            err(str(exc))
            errors = True
    else:
        err(
            "Unable to locate burp-server configuration: {} does not "
            "exist.".format(bconfsrv)
        )
        errors = True

    if errors:
        if not tips:
            err(
                "Some errors have been found in your configuration. "
                "Please make sure you ran this command with the right flags! "
                "(see --help for details)."
            )
        else:
            info(
                "\n"
                "Well, if you are sure about your settings, you can run the "
                "following command to help you setup your Burp-UI agent. "
                "(Note, the '--dry' flag is here to show you the "
                "modifications that will be applied. Once you are OK with "
                "those, you can re-run the command without the '--dry' flag):"
            )
            log(
                '    > bui-manage setup-burp --host="{}" --client="{}" --dry'.format(
                    host, client
                )
            )
    else:
        ok(
            "Congratulations! It seems everything is alright. Burp-UI "
            "should run without any issue now.",
        )


@app.cli.command()
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Dump parts of the config (Please double check no sensitive" " data leaked)",
)
@click.option(
    "-l", "--load", is_flag=True, help="Load all configured modules for full summary"
)
def sysinfo(verbose, load):
    """Returns a couple of system informations to help debugging."""
    from .desc import __release__, __version__
    import platform

    msg = None
    if load:
        try:
            msg = app.load_modules()
        except Exception as e:
            msg = str(e)

    backend = app.config["BACKEND"]

    colors = {
        "True": "green",
        "False": "red",
    }
    embedded_ws = str(app.config["WITH_WS"])
    available_ws = str(WS_AVAILABLE)

    log(
        "Python version:      {}.{}.{}".format(
            sys.version_info[0], sys.version_info[1], sys.version_info[2]
        )
    )
    log("Burp-UI version:     {} ({})".format(__version__, __release__))
    log(
        "OS:                  {}:{} ({})".format(
            platform.system(), platform.release(), os.name
        )
    )
    if platform.system() == "Linux":
        log("Distribution:        {} {} {}".format(*platform.dist()))
    log("Single mode:         {}".format(app.config["STANDALONE"]))
    log("Backend:             {}".format(backend))
    log(
        "WebSocket embedded:  {}".format(
            click.style(embedded_ws, fg=colors[embedded_ws])
        )
    )
    log(
        "WebSocket available: {}".format(
            click.style(available_ws, colors[available_ws])
        )
    )
    log("Config file:         {}".format(app.config.conffile))
    if load:
        if not app.config["STANDALONE"] and not msg:
            log("Agents:")
            for agent, obj in app.client.servers.items():
                client_version = server_version = "unknown"
                try:
                    app.client.status(agent=agent)
                    client_version = app.client.get_client_version(agent=agent)
                    server_version = app.client.get_server_version(agent=agent)
                except BUIserverException:
                    pass
                alive = obj.ping()
                if alive:
                    status = click.style("ALIVE", fg="green")
                else:
                    status = click.style("DISCONNECTED", fg="red")
                log(" - {} ({})".format(agent, status))
                log("   * client version: {}".format(client_version))
                log("   * server version: {}".format(server_version))
        elif not msg:
            server_version = "unknown"
            try:
                app.client.status()
                server_version = app.client.get_server_version()
            except BUIserverException:
                pass
            log("Burp client version: {}".format(app.client.client_version))
            log("Burp server version: {}".format(server_version))
    if verbose:
        log(">>>>> Extra verbose informations:")
        err("!!! PLEASE MAKE SURE NO SENSITIVE DATA GET EXPOSED !!!", err=False)
        sections = [
            "WebSocket",
            "Burp",
            "Production",
            "Global",
        ]
        sections.reverse()
        for section in sections:
            if section in app.config.options:
                log()
                log("    8<{}BEGIN[{}]".format("-" * (67 - len(section)), section))
                for key, val in app.config.options.get(section, {}).items():
                    log("    {} = {}".format(key, val))
                log("    8<{}END[{}]".format("-" * (69 - len(section)), section))

    if load and msg:
        _die(msg, "sysinfo")



# -*- coding: utf8 -*-
"""
.. module:: burpui.decorators
    :platform: Unix
    :synopsis: Burp-UI decorators module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import time
import datetime

from functools import wraps


def browser_cache(expires=None):
    """Add Flask cache response headers based on expires in seconds.

    If expires is None, caching will be disabled.
    Otherwise, caching headers are set to expire in now + expires seconds

    Example usage:

    ::

        @app.route('/map')
        @browser_cache(expires=60)
        def index():
            return render_template('index.html')

    """
    from wsgiref.handlers import format_date_time
    from flask import g
    from flask_restx.utils import unpack

    def cache_decorator(view):
        @wraps(view)
        def cache_func(*args, **kwargs):

            resp, code, headers = unpack(view(*args, **kwargs))
            now = datetime.datetime.now()

            headers["Last-Modified"] = format_date_time(time.mktime(now.timetuple()))

            failure = code - 200 >= 100  # this is not a successful answer
            do_not_cache = getattr(g, "DONOTCACHE", False)

            if expires is None or failure or do_not_cache:
                headers[
                    "Cache-Control"
                ] = "no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0"
            else:
                headers["Cache-Control"] = "private, max-age={}".format(expires)

            return resp, code, headers

        return cache_func

    return cache_decorator


def implement(func):
    """A decorator indicating the method is implemented.

    For the agent and the 'multi' backend, we inherit the backend interface but
    we don't really implement it because we just act as a proxy.
    But maintaining the exhaustive list of methods in several places to always
    implement the same "proxy" thing was painful so I ended up cheating to
    dynamically implement those methods thanks to the __getattribute__ magic
    function.

    But sometimes we want to implement specific things, hence this decorator
    to indicate we don't want the default "magic" implementation and use the
    custom implementation instead.
    """
    try:
        func.__ismethodimplemented__ = True
    except AttributeError:
        # properties seem immutable
        pass
    return func


def usetriorun(func):
    """A decorator indicating the method uses trio.run

    Such functions should always be written like this:

    ::
        def function(self, *args, **kwargs):
            return trio.run(self._async_function, partial(*args, **kwargs))
    """
    try:
        func.__isusingtriorun__ = True
    except AttributeError:
        pass
    return func



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.client
    :platform: Unix
    :synopsis: Burp-UI client api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re

from . import api, cache_key, force_refresh
from ..engines.server import BUIServer  # noqa
from .custom import fields, Resource
from ..decorators import browser_cache
from ..ext.cache import cache
from ..exceptions import BUIserverException

from flask_restx.marshalling import marshal
from flask_restx import inputs
from flask import current_app, request
from flask_login import current_user

bui = current_app  # type: BUIServer
ns = api.namespace("client", "Client methods")

node_fields = ns.model(
    "ClientTree",
    {
        "date": fields.DateTime(
            required=True,
            dt_format="iso8601",
            description="Human representation of the backup date",
        ),
        "gid": fields.Integer(required=True, description="gid owner of the node"),
        "inodes": fields.Integer(required=True, description="Inodes of the node"),
        "mode": fields.String(
            required=True, description='Human readable mode. Example: "drwxr-xr-x"'
        ),
        "name": fields.String(required=True, description="Node name"),
        "title": fields.SafeString(
            required=True, description="Node name (alias)", attribute="name"
        ),
        "fullname": fields.String(required=True, description="Full name of the Node"),
        "key": fields.String(
            required=True,
            description="Full name of the Node (alias)",
            attribute="fullname",
        ),
        "parent": fields.String(required=True, description="Parent node name"),
        "size": fields.String(
            required=True, description='Human readable size. Example: "12.0KiB"'
        ),
        "type": fields.String(required=True, description='Node type. Example: "d"'),
        "uid": fields.Integer(required=True, description="uid owner of the node"),
        "selected": fields.Boolean(
            required=False, description="Is path selected", default=False
        ),
        "lazy": fields.Boolean(
            required=False,
            description="Do the children have been loaded during this"
            + " request or not",
            default=True,
        ),
        "folder": fields.Boolean(required=True, description="Is it a folder"),
        "expanded": fields.Boolean(
            required=False, description="Should we expand the node", default=False
        ),
        # Cannot use nested on own
        "children": fields.Raw(required=False, description="List of children"),
    },
)


@ns.route(
    "/browse/<name>/<int:backup>",
    "/<server>/browse/<name>/<int:backup>",
    endpoint="client_tree",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in" + " multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class ClientTree(Resource):
    """The :class:`burpui.api.client.ClientTree` resource allows you to
    retrieve a list of files in a given backup.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    A mandatory ``GET`` parameter called ``root`` is used to know what path we
    are working on.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument(
        "root",
        help="Root path to expand. You may specify several of them",
        action="append",
    )
    parser.add_argument(
        "recursive",
        type=inputs.boolean,
        help="Returns the whole tree instead of just the sub-tree",
        nullable=True,
        required=False,
        default=False,
    )
    parser.add_argument(
        "selected",
        type=inputs.boolean,
        help="Make the returned path selected at load time. Only works"
        + " if 'recursive' is True",
        nullable=True,
        required=False,
        default=False,
    )
    parser.add_argument(
        "init",
        type=inputs.boolean,
        help="First call to load the root of the tree",
        nullable=True,
        required=False,
        default=False,
    )

    @cache.cached(timeout=3600, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(node_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    @browser_cache(3600)
    def get(self, server=None, name=None, backup=None):
        """Returns a list of 'nodes' under a given path

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "date": "2015-05-21 14:54:49",
                "gid": "0",
                "inodes": "173",
                "selected": false,
                "expanded": false,
                "children": [],
                "mode": "drwxr-xr-x",
                "name": "/",
                "key": "/",
                "title": "/",
                "fullname": "/",
                "parent": "",
                "size": "12.0KiB",
                "type": "d",
                "uid": "0"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int

        :returns: The *JSON* described above.
        """
        args = self.parser.parse_args()
        server = server or args["serverName"]
        json = []
        if not name or not backup:
            return json

        root_list = sorted(args["root"]) if args["root"] else []
        root_loaded = False
        paths_loaded = []
        to_select_list = []

        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(name, server)
        ):
            self.abort(403, "Sorry, you are not allowed to view this client")

        from_cookie = None
        if args["init"] and not root_list:
            from_cookie = request.cookies.get("fancytree-1-expanded", "")
            if from_cookie:
                args["recursive"] = True
                _root = bui.client.get_tree(name, backup, agent=server)
                root_list = [x["name"] for x in _root]
                for path in from_cookie.split("~"):
                    if not path.endswith("/"):
                        path += "/"
                    if path not in root_list:
                        root_list.append(path)
                root_list = sorted(root_list)

        try:
            root_list_clean = []

            for root in root_list:
                if args["recursive"]:
                    path = ""
                    # fetch the root first if not already loaded
                    if not root_loaded:
                        part = bui.client.get_tree(name, backup, level=0, agent=server)
                        root_loaded = True
                    else:
                        part = []
                    root = root.rstrip("/")
                    to_select = root.rsplit("/", 1)
                    if not to_select[0]:
                        to_select[0] = "/"
                    if len(to_select) == 1:
                        # special case we want to select '/'
                        to_select = ("", "/")
                    if not root:
                        root = "/"
                    to_select_list.append(to_select)
                    root_list_clean.append(root)
                    paths = root.split("/")
                    for level, sub in enumerate(paths, start=1):
                        path = os.path.join(path, sub)
                        if not path:
                            path = "/"
                        if path in paths_loaded:
                            continue
                        temp = bui.client.get_tree(
                            name, backup, path, level, agent=server
                        )
                        paths_loaded.append(path)
                        part += temp
                else:
                    part = bui.client.get_tree(name, backup, root, agent=server)
                json += part

            if args["selected"]:
                for entry in json:
                    for parent, fold in to_select_list:
                        if entry["parent"] == parent and entry["name"] == fold:
                            entry["selected"] = True
                            break
                    if entry["parent"] in root_list_clean:
                        entry["selected"] = True

            if not root_list:
                json = bui.client.get_tree(name, backup, agent=server)
                if args["selected"]:
                    for entry in json:
                        if not entry["parent"]:
                            entry["selected"] = True

            if args["recursive"]:
                tree = {}
                rjson = []
                roots = []
                for entry in json:
                    # /!\ after marshalling, 'fullname' will be 'key'
                    tree[entry["fullname"]] = marshal(entry, node_fields)

                for key, entry in tree.items():
                    parent = entry["parent"]
                    if not entry["children"]:
                        entry["children"] = None
                    if parent:
                        node = tree[parent]
                        if not node["children"]:
                            node["children"] = []
                        node["children"].append(entry)
                        if node["folder"]:
                            node["lazy"] = False
                            node["expanded"] = True
                    else:
                        roots.append(entry["key"])

                for fullname in roots:
                    rjson.append(tree[fullname])

                json = rjson
            else:
                for entry in json:
                    entry["children"] = None
                    if not entry["folder"]:
                        entry["lazy"] = False

        except BUIserverException as e:
            self.abort(500, str(e))
        return json


@ns.route(
    "/browseall/<name>/<int:backup>",
    "/<server>/browseall/<name>/<int:backup>",
    endpoint="client_tree_all",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in" + " multi-agent mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class ClientTreeAll(Resource):
    """The :class:`burpui.api.client.ClientTreeAll` resource allows you to
    retrieve a list of all the files in a given backup.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )

    @cache.cached(timeout=3600, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(node_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "405": "Method not allowed",
            "500": "Internal failure",
        },
    )
    @browser_cache(3600)
    def get(self, server=None, name=None, backup=None):
        """Returns a list of all 'nodes' of a given backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "date": "2015-05-21 14:54:49",
                "gid": "0",
                "inodes": "173",
                "selected": false,
                "expanded": false,
                "children": [],
                "mode": "drwxr-xr-x",
                "name": "/",
                "key": "/",
                "title": "/",
                "fullname": "/",
                "parent": "",
                "size": "12.0KiB",
                "type": "d",
                "uid": "0"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int

        :returns: The *JSON* described above.
        """
        args = self.parser.parse_args()
        server = server or args["serverName"]

        if not bui.client.get_attr("batch_list_supported", False, server):
            self.abort(405, "Sorry, the requested backend does not support this method")

        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(name, server)
        ):
            self.abort(403, "Sorry, you are not allowed to view this client")

        try:
            json = self._get_tree_all(name, backup, server)
        except BUIserverException as e:
            self.abort(500, str(e))
        return json

    @staticmethod
    def _get_tree_all(name, backup, server):
        json = bui.client.get_tree(name, backup, "*", agent=server)
        tree = {}
        rjson = []
        roots = []

        def __expand_json(js):
            res = {}
            for entry in js:
                # /!\ after marshalling, 'fullname' will be 'key'
                res[entry["fullname"]] = marshal(entry, node_fields)
            return res

        tree = __expand_json(json)

        # TODO: we can probably improve this at some point
        redo = True
        while redo:
            redo = False
            # using a copy
            for key, entry in dict(tree).items():
                parent = entry["parent"]
                if not entry["children"]:
                    entry["children"] = None
                if parent:
                    if parent not in tree:
                        parent2 = parent
                        last = False
                        while parent not in tree and not last:
                            if not parent2:
                                last = True
                            json = bui.client.get_tree(
                                name, backup, parent2, agent=server
                            )
                            if parent2 == "/":
                                parent2 = ""
                            else:
                                parent2 = os.path.dirname(parent2)
                            tree2 = __expand_json(json)
                            tree.update(tree2)
                        roots = []
                        redo = True
                        break
                    node = tree[parent]
                    if not node["children"]:
                        node["children"] = []
                    elif entry in node["children"]:
                        continue
                    node["children"].append(entry)
                    if node["folder"]:
                        node["lazy"] = False
                        node["expanded"] = False
                else:
                    roots.append(entry["key"])

        for fullname in roots:
            rjson.append(tree[fullname])

        return rjson


@ns.route(
    "/report/<name>",
    "/<server>/report/<name>",
    "/report/<name>/<int:backup>",
    "/<server>/report/<name>/<int:backup>",
    endpoint="client_report",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent" + " mode",
        "name": "Client name",
        "backup": "Backup number",
    },
)
class ClientReport(Resource):
    """The :class:`burpui.api.client.ClientStats` resource allows you to
    retrieve a report on a given backup for a given client.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    report_tpl_fields = ns.model(
        "ClientReportTpl",
        {
            "changed": fields.Integer(
                required=True, description="Number of changed files", default=0
            ),
            "deleted": fields.Integer(
                required=True, description="Number of deleted files", default=0
            ),
            "new": fields.Integer(
                required=True, description="Number of new files", default=0
            ),
            "scanned": fields.Integer(
                required=True, description="Number of scanned files", default=0
            ),
            "total": fields.Integer(
                required=True, description="Total number of files", default=0
            ),
            "unchanged": fields.Integer(
                required=True, description="Number of scanned files", default=0
            ),
        },
    )
    report_fields = ns.model(
        "ClientReport",
        {
            "dir": fields.Nested(report_tpl_fields, required=True),
            "duration": fields.Integer(
                required=True, description="Backup duration in seconds"
            ),
            "efs": fields.Nested(report_tpl_fields, required=True),
            "encrypted": fields.Boolean(
                required=True, description="Is the backup encrypted"
            ),
            "end": fields.DateTime(
                dt_format="iso8601",
                required=True,
                description="Timestamp of the end date of the backup",
            ),
            "files": fields.Nested(report_tpl_fields, required=True),
            "files_enc": fields.Nested(report_tpl_fields, required=True),
            "hardlink": fields.Nested(report_tpl_fields, required=True),
            "meta": fields.Nested(report_tpl_fields, required=True),
            "meta_enc": fields.Nested(report_tpl_fields, required=True),
            "number": fields.Integer(required=True, description="Backup number"),
            "received": fields.Integer(required=True, description="Bytes received"),
            "softlink": fields.Nested(report_tpl_fields, required=True),
            "special": fields.Nested(report_tpl_fields, required=True),
            "start": fields.DateTime(
                dt_format="iso8601",
                required=True,
                description="Timestamp of the beginning of the backup",
            ),
            "totsize": fields.Integer(
                required=True, description="Total size of the backup"
            ),
            "vssfooter": fields.Nested(report_tpl_fields, required=True),
            "vssfooter_enc": fields.Nested(report_tpl_fields, required=True),
            "vssheader": fields.Nested(report_tpl_fields, required=True),
            "vssheader_enc": fields.Nested(report_tpl_fields, required=True),
            "windows": fields.Boolean(
                required=True, description="Is the client a windows system"
            ),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(report_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None, name=None, backup=None):
        """Returns a global report of a given backup/client

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "dir": {
                "changed": 0,
                "deleted": 0,
                "new": 394,
                "scanned": 394,
                "total": 394,
                "unchanged": 0
              },
              "duration": 5,
              "efs": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "encrypted": true,
              "end": 1422189124,
              "files": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "files_enc": {
                "changed": 0,
                "deleted": 0,
                "new": 1421,
                "scanned": 1421,
                "total": 1421,
                "unchanged": 0
              },
              "hardlink": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "meta": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "meta_enc": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "number": 1,
              "received": 1679304,
              "softlink": {
                "changed": 0,
                "deleted": 0,
                "new": 1302,
                "scanned": 1302,
                "total": 1302,
                "unchanged": 0
              },
              "special": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "start": 1422189119,
              "total": {
                "changed": 0,
                "deleted": 0,
                "new": 3117,
                "scanned": 3117,
                "total": 3117,
                "unchanged": 0
              },
              "totsize": 5345361,
              "vssfooter": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "vssfooter_enc": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "vssheader": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "vssheader_enc": {
                "changed": 0,
                "deleted": 0,
                "new": 0,
                "scanned": 0,
                "total": 0,
                "unchanged": 0
              },
              "windows": "false"
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()["serverName"]
        json = []
        if not name:
            err = [[1, "No client defined"]]
            self.abort(400, err)

        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and not current_user.acl.is_client_allowed(name, server)
        ):
            self.abort(403, "You don't have rights to view this client report")
        if backup:
            try:
                json = bui.client.get_backup_logs(backup, name, agent=server)
            except BUIserverException as exp:
                self.abort(500, str(exp))
        else:
            try:
                json = bui.client.get_backup_logs(-1, name, agent=server)
            except BUIserverException as exp:
                self.abort(500, str(exp))
        return json

    @api.disabled_on_demo()
    @ns.marshal_with(report_fields, code=202, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "400": "Missing arguments",
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    def delete(self, name, backup, server=None):
        """Deletes a given backup from the server

        **DELETE** method provided by the webservice.

        The access is filtered by the :mod:`burpui.misc.acl` module so that you
        can only delete backups you have access to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        server = server or self.parser.parse_args()["serverName"]
        if not name:
            err = [[1, "No client defined"]]
            self.abort(400, err)

        if (
            not current_user.is_anonymous
            and not current_user.acl.is_admin()
            and (
                not current_user.acl.is_moderator()
                or current_user.acl.is_moderator()
                and not current_user.acl.is_client_rw(name, server)
            )
        ):
            self.abort(403, "You don't have rights on this client")

        msg = bui.client.delete_backup(name, backup, server)
        if msg:
            self.abort(500, msg)
        bui.audit.logger.info(
            f"requested the deletion of backup {backup} for {name}", server=server
        )
        return 202, ""


@ns.route("/stats/<name>", "/<server>/stats/<name>", endpoint="client_stats")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent" + " mode",
        "name": "Client name",
    },
)
class ClientStats(Resource):
    """The :class:`burpui.api.client.ClientReport` resource allows you to
    retrieve a list of backups for a given client.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    client_fields = ns.model(
        "ClientStats",
        {
            "number": fields.Integer(required=True, description="Backup number"),
            "received": fields.Integer(required=True, description="Bytes received"),
            "size": fields.Integer(required=True, description="Total size"),
            "encrypted": fields.Boolean(
                required=True, description="Is the backup encrypted"
            ),
            "deletable": fields.Boolean(
                required=True, description="Is the backup deletable"
            ),
            "date": fields.DateTime(
                required=True,
                dt_format="iso8601",
                description="Human representation of the backup date",
            ),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(client_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None, name=None):
        """Returns a list of backups for a given client

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "date": "2015-01-25 13:32:00",
                "deletable": true,
                "encrypted": true,
                "received": 123,
                "size": 1234,
                "number": 1
              },
            ]

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()["serverName"]
        try:
            if (
                not current_user.is_anonymous
                and not current_user.acl.is_admin()
                and not current_user.acl.is_client_allowed(name, server)
            ):
                self.abort(403, "Sorry, you cannot access this client")
            json = bui.client.get_client(name, agent=server)
        except BUIserverException as exp:
            self.abort(500, str(exp))
        return json


@ns.route("/labels/<name>", "/<server>/labels/<name>", endpoint="client_labels")
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent" + " mode",
        "name": "Client name",
    },
)
class ClientLabels(Resource):
    """The :class:`burpui.api.client.ClientLabels` resource allows you to
    retrieve the labels of a given client.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument("clientName", help="Client name")
    labels_fields = ns.model(
        "ClientLabels",
        {
            "labels": fields.List(fields.String, description="List of labels"),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(labels_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self, server=None, name=None):
        """Returns the labels of a given client

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "labels": [
                "label1",
                "label2"
              ]
            }

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: The *JSON* described above.
        """
        try:
            if (
                not current_user.is_anonymous
                and not current_user.acl.is_admin()
                and not current_user.acl.is_client_allowed(name, server)
            ):
                self.abort(403, "Sorry, you cannot access this client")
            labels = self._get_labels(name, server)
        except BUIserverException as exp:
            self.abort(500, str(exp))
        return {"labels": labels}

    @staticmethod
    def _get_labels(client, server=None):
        key = "labels-{}-{}".format(client, server)
        ret = cache.cache.get(key)
        if ret is not None:
            return ret
        labels = bui.client.get_client_labels(client, agent=server)
        ret = []
        ignore = re.compile("|".join(bui.ignore_labels)) if bui.ignore_labels else None
        reformat = (
            [(re.compile(regex), replace) for regex, replace in bui.format_labels]
            if bui.format_labels
            else []
        )
        for label in labels:
            if bui.ignore_labels and ignore.search(label):
                continue
            tmp_label = label
            for regex, replace in reformat:
                tmp_label = regex.sub(replace, tmp_label)
            ret.append(tmp_label)
        cache.cache.set(key, ret, 1800)
        return ret


@ns.route(
    "/running",
    "/running/<name>",
    "/<server>/running",
    "/<server>/running/<name>",
    endpoint="client_running_status",
)
@ns.doc(
    params={
        "server": "Which server to collect data from when in multi-agent" + " mode",
        "name": "Client name",
    },
)
class ClientRunningStatus(Resource):
    """The :class:`burpui.api.client.ClientRunningStatus` resource allows you to
    retrieve the running status of a given client.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    parser = ns.parser()
    parser.add_argument(
        "serverName", help="Which server to collect data from when in multi-agent mode"
    )
    parser.add_argument("clientName", help="Client name")
    running_fields = ns.model(
        "ClientRunningStatus",
        {
            "state": fields.LocalizedString(required=True, description="Running state"),
            "percent": fields.Integer(
                required=False, description="Backup progress in percent", default=-1
            ),
            "phase": fields.String(
                required=False, description="Backup phase", default=None
            ),
            "last": fields.DateTime(
                required=False, dt_format="iso8601", description="Date of last backup"
            ),
        },
    )

    @ns.marshal_list_with(running_fields, code=200, description="Success")
    @ns.expect(parser)
    @ns.doc(
        responses={
            "403": "Insufficient permissions",
            "500": "Internal failure",
        },
    )
    def get(self, server=None, name=None):
        """Returns the running status of a given client

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "state": "running",
              "percent": 42,
              "phase": "2",
              "last": "now"
            }

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :returns: The *JSON* described above.
        """
        args = self.parser.parse_args()
        server = server or args["serverName"]
        name = name or args["clientName"]
        try:
            if (
                not current_user.is_anonymous
                and not current_user.acl.is_admin()
                and not current_user.acl.is_client_allowed(name, server)
            ):
                self.abort(403, "Sorry, you cannot access this client")
            json = bui.client.get_client_status(name, agent=server)
        except BUIserverException as exp:
            self.abort(500, str(exp))
        return json



# -*- coding: utf8 -*-
"""
.. module:: burpui.routes
    :platform: Unix
    :synopsis: Burp-UI routes module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import re
import math
import uuid

from flask import (
    request,
    render_template,
    redirect,
    url_for,
    abort,
    flash,
    Blueprint,
    session,
    current_app,
    g,
)
from flask_login import login_user, login_required, logout_user, current_user
from flask_babel import gettext as _, refresh as refresh_babel

from .desc import __url__, __doc__
from .engines.server import BUIServer  # noqa
from .sessions import session_manager
from ._compat import quote
from .forms import LoginForm
from .exceptions import BUIserverException
from .utils import human_readable as _hr
from .security import sanitize_string, is_safe_url


bui = current_app  # type: BUIServer
view = Blueprint("view", "burpui", template_folder="templates")


"""
Here are some custom filters
"""


@view.app_template_filter()
def mypad(value):
    """
    Filter: used to pad 0's to backup numbers as in the burp's status monitor
    """
    if not value:
        return "0000000"
    return "{0:07d}".format(int(value))


@view.app_template_filter()
def time_human(value):
    """Convert the input value to human readable time"""
    string = ""
    seconds = (((value % 31536000) % 86400) % 3600) % 60
    minutes = math.floor((((value % 31536000) % 86400) % 3600) / 60)
    hours = math.floor(((value % 31536000) % 86400) / 3600)
    if hours > 0:
        string = "%02dH" % hours
    return "%s %02dm %02ds" % (string, minutes, seconds)


@view.app_template_filter()
def bytes_human(value):
    """Convert the input value to human readable bytes"""
    return "{0:.1eM}".format(_hr(value))


@view.app_template_filter()
def regex_replace(value, regex, replace):
    """Replace every string matching the given regex with the replacement"""
    return re.sub(regex, replace, value, flags=re.IGNORECASE)


"""
And here is the main site
"""


@view.route("/calendar")
@view.route("/calendar/<client>")
@view.route("/<server>/calendar")
@view.route("/<server>/calendar/<client>")
@login_required
def calendar(server=None, client=None):
    server = server or request.args.get("serverName")
    client = client or request.args.get("clientName")
    if not bui.config["STANDALONE"]:
        servers = not server and not client
        clients = bool(server)
        cli = bool(client)
    else:
        servers = False
        cli = bool(client)
        clients = not cli
    return render_template(
        "calendar.html",
        calendar=True,
        server=server,
        cname=client,
        client=cli,
        servers=servers,
        clients=clients,
    )


@view.route("/settings")
@view.route("/settings/<path:conf>")
@view.route("/<server>/settings")
@view.route("/<server>/settings/<path:conf>")
@login_required
def settings(server=None, conf=None):
    # Only the admin can edit the configuration
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    if not conf:
        try:
            conf = quote(request.args.get("conf"), safe="")
            if conf:
                return redirect(url_for(".settings", server=server, conf=conf))
        except:
            pass
    server = server or request.args.get("serverName")
    return render_template(
        "settings.html",
        settings=True,
        is_admin=current_user.acl.is_admin(),
        is_moderator=current_user.acl.is_moderator(),
        server=server,
        conf=conf,
        ng_controller="ConfigCtrl",
    )


@view.route("/admin/sessions/<user>")
@login_required
def admin_sessions(user):
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    return render_template(
        "admin/sessions.html",
        admin=True,
        sessions=True,
        user=user,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/authentication/<user>")
@login_required
def admin_authentication(user):
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    backend = request.args.get("backend")
    if not backend:
        abort(400)
    return render_template(
        "admin/authentication.html",
        admin=True,
        authentication=True,
        user=user,
        backend=backend,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/authorization/grant/<grant>")
@login_required
def admin_grant_authorization(grant):
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    backend = request.args.get("backend")
    if not backend:
        abort(400)
    return render_template(
        "admin/grant-authorization.html",
        admin=True,
        authorization=True,
        grant=grant,
        backend=backend,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/authorization/group/<group>")
@login_required
def admin_group_authorization(group):
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    backend = request.args.get("backend")
    if not backend:
        abort(400)
    return render_template(
        "admin/group-authorization.html",
        admin=True,
        authorization=True,
        group=group,
        backend=backend,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/backend/<backend>")
@login_required
def admin_backend(backend):
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    return render_template(
        "admin/backend.html",
        admin=True,
        onebackend=True,
        backend=backend,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/backends")
@login_required
def admin_backends():
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    return render_template(
        "admin-backends.html", admin=True, backends=True, ng_controller="AdminCtrl"
    )


@view.route("/admin/authorizations")
@login_required
def admin_authorizations():
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    return render_template(
        "admin-authorizations.html",
        admin=True,
        authorizations=True,
        ng_controller="AdminCtrl",
    )


@view.route("/admin/authentications")
@login_required
def admin_authentications():
    # Only the admin can access this page
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    return render_template(
        "admin-authentications.html",
        admin=True,
        authentications=True,
        ng_controller="AdminCtrl",
    )


@view.route("/me")
@login_required
def me():
    return render_template("user.html", me=True, ng_controller="UserCtrl")


@view.route("/client/client-settings")
@view.route("/<client>/client-settings")
@view.route("/<client>/client-settings/<path:conf>")
@view.route("/<server>/client/client-settings")
@view.route("/<server>/<client>/client-settings")
@view.route("/<server>/<client>/client-settings/<path:conf>")
@login_required
def cli_settings(server=None, client=None, conf=None):
    # Only the admin can edit the configuration
    if (
        not current_user.is_anonymous
        and not current_user.acl.is_admin()
        and not current_user.acl.is_moderator()
    ):
        abort(403)
    if not conf:
        try:
            conf = quote(request.args.get("conf"), safe="")
            if conf:
                return redirect(
                    url_for(".cli_settings", server=server, client=client, conf=conf)
                )
        except:
            pass
    client = client or request.args.get("client")
    server = server or request.args.get("serverName")
    template = request.args.get("template") or False
    statictemplate = request.args.get("statictemplate") or False
    return render_template(
        "settings.html",
        settings=True,
        client_mode=True,
        template=template,
        statictemplate=statictemplate,
        client=client,
        server=server,
        conf=conf,
        ng_controller="ConfigCtrl",
    )


@view.route("/live-monitor")
@view.route("/<server>/live-monitor")
@view.route("/live-monitor/<name>")
@view.route("/<server>/live-monitor/<name>")
@login_required
def live_monitor(server=None, name=None):
    """Live status monitor view"""
    server = server or request.args.get("serverName")

    return render_template(
        "live-monitor.html",
        live=True,
        cname=name,
        server=server,
        ng_controller="LiveCtrl",
    )


@view.route("/edit-server-initiated-restore/<name>", methods=["GET"])
@view.route("/<server>/edit-server-initiated-restore/<name>", methods=["GET"])
@login_required
def edit_server_initiated_restore(server=None, name=None):
    data = bui.client.is_server_restore(name, server)
    to = None
    if not data or not data["found"]:
        flash(_("Sorry, there are no restore file found for this client"), "warning")
        return redirect(url_for(".home"))
    if data.get("orig_client"):
        to = name
        name = data["orig_client"]
    return redirect(
        url_for(
            ".client_browse",
            server=server,
            name=name,
            backup=data["backup"],
            edit=1,
            to=to,
        )
    )


@view.route("/client-browse/<name>", methods=["GET"])
@view.route("/<server>/client-browse/<name>", methods=["GET"])
@view.route("/client-browse/<name>/<int:backup>")
@view.route("/<server>/client-browse/<name>/<int:backup>")
@view.route("/client-browse/<name>/<int:backup>/<int:encrypted>")
@view.route("/<server>/client-browse/<name>/<int:backup>/<int:encrypted>")
@login_required
def client_browse(server=None, name=None, backup=None, encrypted=None, edit=None):
    """Browse a specific backup of a specific client"""
    if request.args.get("encrypted") == "1":
        encrypted = 1
    if request.args.get("edit") == "1":
        to = request.args.get("to") or name
        edit = bui.client.is_server_restore(to, server)
        if not edit or not edit["found"]:
            flash(
                _("Sorry, there are no restore file found for this client"), "warning"
            )
            edit = None
        else:
            edit["roots"] = [x["key"] for x in edit["list"]]
    server = server or request.args.get("serverName")
    bkp = request.args.get("backup")
    if bkp and not backup:
        return redirect(
            url_for(
                ".client_browse",
                name=name,
                backup=bkp,
                encrypted=encrypted,
                server=server,
            )
        )
    return render_template(
        "client-browse.html",
        tree=True,
        backup=True,
        overview=True,
        cname=name,
        nbackup=backup,
        encrypted=encrypted,
        server=server,
        edit=edit,
        ng_controller="BrowseCtrl",
    )


@view.route("/client-report/<name>")
@view.route("/<server>/client-report/<name>")
@login_required
def client_report(server=None, name=None):
    """Specific client report"""
    server = server or request.args.get("serverName")
    try:
        res = bui.client.get_client(name, agent=server)
    except BUIserverException:
        res = []
    if len(res) == 1:
        return redirect(
            url_for(".backup_report", name=name, backup=res[0]["number"], server=server)
        )
    return render_template(
        "client-report.html", client=True, report=True, cname=name, server=server
    )


@view.route("/clients-report")
@view.route("/<server>/clients-report")
@login_required
def clients_report(server=None):
    """Global report"""
    server = server or request.args.get("serverName")
    return render_template(
        "clients-report.html", clients=True, report=True, server=server
    )


@view.route("/backup-report/<name>", methods=["GET"])
@view.route("/<server>/backup-report/<name>", methods=["GET"])
@view.route("/backup-report/<name>/<int:backup>", methods=["GET"])
@view.route("/<server>/backup-report/<name>/<int:backup>", methods=["GET"])
@login_required
def backup_report(server=None, name=None, backup=None):
    """Backup specific report"""
    backup = backup or request.args.get("backup")
    server = server or request.args.get("serverName")
    return render_template(
        "backup-report.html",
        client=True,
        backup=True,
        report=True,
        cname=name,
        nbackup=backup,
        server=server,
    )


@view.route("/client", methods=["GET"])
@view.route("/<server>/client", methods=["GET"])
@view.route("/client/<name>")
@view.route("/<server>/client/<name>")
@login_required
def client(server=None, name=None):
    """Specific client overview"""
    c = name or request.args.get("name")
    server = server or request.args.get("serverName")
    # burp 2 allows to browse backups will backing up whereas burp 1 does not
    if bui.client.version(server) == 1 and bui.client.is_backup_running(
        c, agent=server
    ):
        return redirect(url_for(".live_monitor", name=c, server=server))
    return render_template(
        "client.html", client=True, overview=True, cname=c, server=server
    )


@view.route("/clients", methods=["GET"])
@view.route("/<server>/clients", methods=["GET"])
@login_required
def clients(server=None):
    server = server or request.args.get("serverName")
    return render_template("clients.html", clients=True, overview=True, server=server)


@view.route("/servers", methods=["GET"])
@login_required
def servers():
    return render_template("servers.html", servers=True, overview=True)


@view.route("/servers-report", methods=["GET"])
@login_required
def servers_report():
    return render_template("servers-report.html", servers=True, report=True)


@view.route("/login", methods=["POST", "GET"])
def login():
    form = LoginForm(request.form)

    if form.validate_on_submit():
        # allow to switch to another backend
        refresh = True
        session["tag_id"] = uuid.uuid4()
        session["language"] = form.language.data
        session["_extra"] = g.now
        user = bui.uhandler.user(form.username.data, refresh)
        # at the time the context is loaded, the locale is not set
        refresh_babel()
        if user and user.is_active and user.login(form.password.data):
            login_user(user, remember=form.remember.data)
            flash(_("Logged in successfully"), "success")
            session_manager.store_session(
                form.username.data,
                request.remote_addr,
                request.headers.get("User-Agent"),
                form.remember.data,
            )
            next_hop = request.args.get("next") or url_for(".home")
            if is_safe_url(next_hop):
                return redirect(next_hop)
            return abort(400)
        else:
            flash(_("Wrong username or password"), "danger")
            sanitized_user = repr(sanitize_string(form.username.data, paranoid=True))
            msg = f"Wrong username or password attempted by {sanitized_user} at {request.remote_addr}"
            if user:
                audit = f"Wrong password for {sanitized_user} at {request.remote_addr}"
            else:
                audit = f"Unknown username {sanitized_user} at {request.remote_addr}"
            bui.logger.critical(msg)
            bui.audit.logger.critical(audit)
    elif form.is_submitted():
        flash(_("Wrong CSRF token, please try again"), "warning")
    return render_template("login.html", form=form, login=True)


@view.route("/logout")
@login_required
def logout():
    session_manager.delete_session()
    # cleanup the session at logout to avoid further reuse by another user
    session.clear()
    logout_user()
    return redirect(url_for(".home"))


@view.route("/about")
def about():
    """about view"""
    return render_template(
        "about.html",
        about=True,
        login=(not current_user.is_authenticated),
        doc=__doc__,
        url=__url__,
    )


@view.route("/")
@login_required
def home():
    """Home page"""
    if bui.config["STANDALONE"]:
        return redirect(url_for(".clients"))
    else:
        server = request.args.get("serverName")
        if server:
            return redirect(url_for(".clients", server=server))
        return redirect(url_for(".servers"))



"""handle sessions

Revision ID: 7f317474332d
Revises: 225d9b2f0fb1
Create Date: 2016-08-30 11:47:35.513396

"""

# revision identifiers, used by Alembic.
revision = "7f317474332d"
down_revision = "225d9b2f0fb1"

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "session",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=256), nullable=True),
        sa.Column("user", sa.String(length=256), nullable=True),
        sa.Column("ip", sa.String(length=256), nullable=True),
        sa.Column("ua", sa.String(length=2048), nullable=True),
        sa.Column("timestamp", sa.DateTime(), nullable=True),
        sa.Column("expire", sa.DateTime(), nullable=True),
        sa.Column("permanent", sa.Boolean(), nullable=True),
        sa.Column("api", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("uuid"),
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("session")
    ### end Alembic commands ###



# -*- coding: utf8 -*-
"""
.. module:: burpui.filter
    :platform: Unix
    :synopsis: Burp-UI filter module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""


class BUIMask(object):
    """Mask client/servers based on user preferences or user ACL"""

    def init_app(self, app):
        """Initialize the mask

        :param app: Application context
        :type app: :class:`burpui.engines.server.BUIServer`
        """
        self.app = app

    @property
    def is_user_pref(self):
        return self.app.config["WITH_SQL"]

    def has_filters(self, current_user):
        if not current_user.is_anonymous and current_user.acl.is_admin():
            if self.is_user_pref:
                return self.query_user(current_user.name)
            return False
        return True

    def query_user(self, username):
        from .models import Hidden

        return Hidden.query.filter_by(user=username).first()

    def query_hidden(self, username):
        from .models import Hidden

        return (
            Hidden.query.filter_by(user=username)
            .with_entities(Hidden.client, Hidden.server)
            .all()
        )

    def hidden_clients(self, username):
        if not self.is_user_pref:
            return []
        return self.query_hidden(username)

    def hidden_servers(self, username):
        if not self.is_user_pref:
            return []
        hidden = self.query_hidden(username)
        return [server for client, server in hidden if not client]

    def is_client_allowed(self, current_user, client=None, server=None):
        if self.app.auth == "none":
            return True
        if current_user.is_anonymous:
            return False
        if self.has_filters(current_user) and self.is_user_pref:
            hidden = self.hidden_clients(current_user.name)
            if (client, server) in hidden:
                return False
        if current_user.acl.is_admin():
            return True
        return current_user.acl.is_client_allowed(client, server)

    def is_client_rw(self, current_user, client, server=None):
        if self.app.auth == "none":
            return True
        if current_user.is_anonymous:
            return False
        if self.has_filters(current_user) and self.is_user_pref:
            hidden = self.hidden_clients(current_user.name)
            if (client, server) in hidden:
                return False
        if current_user.acl.is_admin():
            return True
        return current_user.acl.is_client_rw(client, server)

    def is_server_allowed(self, current_user, server):
        if self.app.auth == "none":
            return True
        if current_user.is_anonymous:
            return False
        if self.has_filters(current_user) and self.is_user_pref:
            hidden = self.hidden_servers(current_user.name)
            if server in hidden:
                return False
        if current_user.acl.is_admin():
            return True
        return current_user.acl.is_server_allowed(server)

    def is_server_rw(self, current_user, server):
        if self.app.auth == "none":
            return True
        if current_user.is_anonymous:
            return False
        if self.has_filters(current_user) and self.is_user_pref:
            hidden = self.hidden_servers(current_user.name)
            if server in hidden:
                return False
        if current_user.acl.is_admin():
            return True
        return current_user.acl.is_server_rw(server)


mask = BUIMask()



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.__main__
    :platform: Unix
    :synopsis: Burp-UI main module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys

from argparse import ArgumentParser, REMAINDER

ROOT = os.path.dirname(os.path.realpath(__file__))
# Try to load modules from our current env first
sys.path.insert(0, os.path.join(ROOT, ".."))


def parse_args(mode=True, name=None):
    mname = name
    if not name:
        mname = "burp-ui"
    parser = ArgumentParser(prog=mname)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="log",
        help="increase output verbosity (e.g., -vv is more verbose than -v)",
        action="count",
    )
    parser.add_argument(
        "-d", "--debug", dest="debug", help="enable debug mode", action="store_true"
    )
    parser.add_argument(
        "-V",
        "--version",
        dest="version",
        help="print version and exit",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        help="burp-ui configuration file",
        metavar="<CONFIG>",
    )
    parser.add_argument(
        "-l",
        "--logfile",
        dest="logfile",
        help="output logs in defined file",
        metavar="<FILE>",
    )
    parser.add_argument(
        "-i",
        "--migrations",
        dest="migrations",
        help="migrations directory",
        metavar="<MIGRATIONSDIR>",
    )
    parser.add_argument("remaining", nargs=REMAINDER)
    if mode:
        parser.add_argument(
            "-m",
            "--mode",
            dest="mode",
            help="application mode",
            metavar="<agent|server|celery|manage|monitor|legacy>",
        )

    options, unknown = parser.parse_known_args()
    if mode and options.mode and options.mode not in ["celery", "manage", "server"]:
        options = parser.parse_args()
        unknown = []

    if options.version:
        from burpui.desc import __title__, __version__, __release__

        ver = "{}: v{}".format(mname or __title__, __version__)
        if options.log:
            ver = "{} ({})".format(ver, __release__)
        print(ver)
        sys.exit(0)

    return options, unknown


def main():
    """
    Main function
    """
    options, unknown = parse_args(mode=True)

    if not options.mode or options.mode == "server":
        server(options, unknown)
    elif options.mode == "agent":
        agent(options)
    elif options.mode == "celery":
        celery()
    elif options.mode == "manage":
        manage()
    elif options.mode == "monitor":
        monitor(options)
    elif options.mode == "legacy":
        legacy(options, unknown)
    else:
        print("Wrong mode!")
        sys.exit(1)


def server(options=None, unknown=None):
    from burpui.utils import lookup_file

    if unknown is None:
        unknown = []
    if not options:
        options, unknown = parse_args(mode=False)
    env = os.environ

    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        if "BUI_CONFIG" in env:
            conf = env["BUI_CONFIG"]
        else:
            conf = lookup_file()
    check_config(conf)

    if os.path.isdir("burpui"):
        env["FLASK_APP"] = "burpui/cli.py"
    else:
        env["FLASK_APP"] = "burpui.cli"
    env["BUI_CONFIG"] = conf
    env["BUI_VERBOSE"] = str(options.log)
    if options.logfile:
        env["BUI_LOGFILE"] = options.logfile
    if options.debug:
        env["BUI_DEBUG"] = "1"
        env["FLASK_DEBUG"] = "1"
    env["BUI_MODE"] = "server"

    args = ["flask", "run"]
    args += unknown
    args += [x for x in options.remaining if x != "--"]

    os.execvpe(args[0], args, env)


def agent(options=None):
    import trio
    from burpui.engines.agent import BUIAgent as Agent
    from burpui.utils import lookup_file

    if not options:
        options, _ = parse_args(mode=False, name="bui-agent")

    conf = ["buiagent.cfg", "buiagent.sample.cfg"]
    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        conf = lookup_file(conf)
    check_config(conf)

    agent = Agent(conf, options.log, options.logfile)
    trio.run(agent.run)


def monitor(options=None):
    import trio
    from burpui.engines.monitor import MonitorPool
    from burpui.utils import lookup_file

    if not options:
        options, _ = parse_args(mode=False, name="bui-agent")

    conf = ["buimonitor.cfg", "buimonitor.sample.cfg"]
    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        conf = lookup_file(conf)
    check_config(conf)

    monitor = MonitorPool(conf, options.log, options.logfile)
    trio.run(monitor.run)


def celery():
    from burpui.utils import lookup_file

    parser = ArgumentParser("bui-celery")
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        help="burp-ui configuration file",
        metavar="<CONFIG>",
    )
    parser.add_argument(
        "-t", "--type", dest="type", help="celery mode", metavar="<worker|beat|flower>"
    )
    parser.add_argument(
        "-m",
        "--mode",
        dest="mode",
        help="application mode",
        metavar="<agent|server|worker|manage|legacy>",
    )
    parser.add_argument("remaining", nargs=REMAINDER)

    options, unknown = parser.parse_known_args()
    env = os.environ

    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        if "BUI_CONFIG" in env:
            conf = env["BUI_CONFIG"]
        else:
            conf = lookup_file()

    if options.type:
        celery_mode = options.type
    else:
        celery_mode = "worker"

    # make conf path absolute
    if not conf.startswith("/"):
        curr = os.getcwd()
        conf = os.path.join(curr, conf)

    check_config(conf)

    os.chdir(ROOT)

    env["BUI_MODE"] = "celery"
    env["BUI_CONFIG"] = conf

    args = ["celery", "-A", "engines.worker.celery", celery_mode]
    args += unknown
    args += [x for x in options.remaining if x != "--"]

    os.execvpe(args[0], args, env)


def manage():
    from burpui.utils import lookup_file

    parser = ArgumentParser("bui-manage")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="log",
        help="increase output verbosity (e.g., -vv is more verbose than -v)",
        action="count",
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        help="burp-ui configuration file",
        metavar="<CONFIG>",
    )
    parser.add_argument(
        "-i",
        "--migrations",
        dest="migrations",
        help="migrations directory",
        metavar="<MIGRATIONSDIR>",
    )
    parser.add_argument(
        "-m",
        "--mode",
        dest="mode",
        help="application mode",
        metavar="<agent|server|worker|manage|legacy>",
    )
    parser.add_argument(
        "-l",
        "--logfile",
        dest="logfile",
        help="output logs in defined file",
        metavar="<FILE>",
    )
    parser.add_argument("remaining", nargs=REMAINDER)

    options, unknown = parser.parse_known_args()
    env = os.environ

    if options.logfile:
        env["BUI_LOGFILE"] = options.logfile
    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        if "BUI_CONFIG" in env:
            conf = env["BUI_CONFIG"]
        else:
            conf = lookup_file()
    check_config(conf)

    if options.migrations:
        migrations = lookup_file(
            options.migrations, guess=False, directory=True, check=False
        )
    else:
        migrations = lookup_file("migrations", directory=True)

    env["BUI_MODE"] = "manage"
    env["BUI_CONFIG"] = conf
    env["BUI_VERBOSE"] = str(options.log)
    if migrations:
        env["BUI_MIGRATIONS"] = migrations
    if os.path.isdir("burpui") and os.path.isfile("burpui/cli.py"):
        env["FLASK_APP"] = "burpui/cli.py"
    else:
        env["FLASK_APP"] = "burpui.cli"

    args = ["flask"]
    args += unknown
    args += [x for x in options.remaining if x != "--"]

    os.execvpe(args[0], args, env)


def legacy(options=None, unknown=None):
    from burpui.utils import lookup_file

    if unknown is None:
        unknown = []
    if not options:
        options, unknown = parse_args(mode=False, name="burpui-legacy")
    env = os.environ

    if options.config:
        conf = lookup_file(options.config, guess=False)
    else:
        if "BUI_CONFIG" in env:
            conf = env["BUI_CONFIG"]
        else:
            conf = lookup_file()
    check_config(conf)

    env["BUI_MODE"] = "legacy"
    env["BUI_CONFIG"] = conf
    if os.path.isdir("burpui"):
        env["FLASK_APP"] = "burpui/cli.py"
    else:
        env["FLASK_APP"] = "burpui.cli"
    env["BUI_VERBOSE"] = str(options.log)
    if options.logfile:
        env["BUI_LOGFILE"] = options.logfile
    if options.debug:
        env["BUI_DEBUG"] = "1"
        env["FLASK_DEBUG"] = "1"

    args = ["flask", "legacy"]
    args += unknown
    args += [x for x in options.remaining if x != "--"]

    os.execvpe(args[0], args, env)


def check_config(conf):
    if not conf:
        raise IOError("No configuration file found")
    if not os.path.isfile(conf):
        raise IOError("File does not exist: '{0}'".format(conf))


if __name__ == "__main__":
    main()



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.i18n
    :platform: Unix
    :synopsis: Burp-UI external Internationalization module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask import request, session
from flask_babel import Babel
from flask_login import current_user
from ..config import config
from .._compat import to_unicode

babel = Babel()

LANGUAGES = {
    "en": to_unicode("English"),
    "fr": to_unicode("Français"),
    "es": to_unicode("Español"),
    "it": to_unicode("Italiano"),
}
config["LANGUAGES"] = LANGUAGES


@babel.localeselector
def get_locale():
    locale = None
    if current_user and not current_user.is_anonymous:
        locale = getattr(current_user, "language", None)
    elif "language" in session:
        locale = session.get("language", None)
    if locale not in LANGUAGES:
        locale = None
    return locale or request.accept_languages.best_match(config["LANGUAGES"].keys())



# -*- coding: utf8 -*-

# This is a submodule we can also use "from ..api import api"
from . import api, cache_key, force_refresh
from ..engines.server import BUIServer  # noqa
from .custom import fields, Resource
from ..filter import mask
from ..ext.cache import cache
from ..decorators import browser_cache
from ..exceptions import BUIserverException

from flask import current_app, g
from flask_login import current_user

bui = current_app  # type: BUIServer
ns = api.namespace("servers", "Servers methods")


@ns.route("/stats", endpoint="servers_stats")
class ServersStats(Resource):
    """The :class:`burpui.api.servers.ServersStats` resource allows you to
    retrieve statistics about servers/agents.

    This resource is part of the :mod:`burpui.api.servers` module.
    """

    servers_fields = ns.model(
        "Servers",
        {
            "alive": fields.Boolean(
                required=True, description="Is the server reachable"
            ),
            "clients": fields.Integer(
                required=True, description="Number of clients managed by this server"
            ),
            "name": fields.String(required=True, description="Server name"),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_list_with(servers_fields, code=200, description="Success")
    @ns.doc(
        responses={
            500: "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self):
        """Returns a list of servers (agents) with basic stats

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                'alive': true,
                'clients': 2,
                'name': 'burp1'
              },
              {
                'alive': false,
                'clients': 0,
                'name': 'burp2'
              }
            ]


        :returns: The *JSON* described above.
        """

        r = []
        check = False

        if bui.config["STANDALONE"]:
            return r

        if not current_user.is_anonymous and mask.has_filters(current_user):
            check = True

        for serv in bui.client.servers:
            try:
                alive = bui.client.servers[serv].ping()
            except BUIserverException:
                alive = False
                g.DONOTCACHE = True

            try:
                clients = bui.client.servers[serv].get_all_clients(
                    serv, last_attempt=False
                )
            except BUIserverException:
                clients = []

            if check and mask.is_server_allowed(current_user, serv):
                allowed_clients = [
                    x
                    for x in clients
                    if mask.is_client_allowed(current_user, x["name"], serv)
                ]
                r.append(
                    {"name": serv, "clients": len(allowed_clients), "alive": alive}
                )
            elif not check:
                r.append({"name": serv, "clients": len(clients), "alive": alive})

        return r


@ns.route("/report", endpoint="servers_report")
class ServersReport(Resource):
    """The :class:`burpui.api.servers.ServersReport` resource allows you to
    retrieve a report about servers/agents.

    This resource is part of the :mod:`burpui.api.servers` module.
    """

    wild = fields.Wildcard(
        fields.Integer,
        required=False,
        description="Total number of clients of a given OS",
    )
    os_fields = ns.model(
        "OS",
        {
            "*": wild,
        },
    )
    stats_fields = ns.model(
        "ServersStats",
        {
            "total": fields.Integer(
                required=True, description="Number of files", default=0
            ),
            "totsize": fields.Integer(
                required=True,
                description="Total size occupied by all the backups of this server",
                default=0,
            ),
            "os": fields.Nested(os_fields),
        },
    )
    server_fields = ns.model(
        "ServersReport",
        {
            "name": fields.String(required=True, description="Server name"),
            "number": fields.Nested(stats_fields, required=True),
        },
    )
    backup_fields = ns.model(
        "ServersBackup",
        {
            "name": fields.String(required=True, description="Server name"),
            "number": fields.Integer(
                required=True, description="Number of backups on this server", default=0
            ),
        },
    )
    report_fields = ns.model(
        "ServersReportFull",
        {
            "backups": fields.Nested(backup_fields, as_list=True, required=True),
            "servers": fields.Nested(server_fields, as_list=True, required=True),
        },
    )

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(report_fields, code=200, description="Success")
    @ns.doc(
        responses={
            403: "Insufficient permissions",
            500: "Internal failure",
        },
    )
    @browser_cache(1800)
    def get(self):
        """Returns a global report about all the servers managed by Burp-UI

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "AGENT1",
                  "number": 49
                }
              ],
              "servers": [
                {
                  "name": "AGENT1",
                  "number": {
                    "os": {
                      "linux": 4
                    },
                    "total": 349705,
                    "totsize": 119400711726
                  }
                }
              ]
            }

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients/servers you are authorized to.

        :returns: The *JSON* described above.
        """
        r = {}
        check = False
        if not current_user.is_anonymous and mask.has_filters(current_user):
            check = True

        backups = []
        servers = []
        try:
            for serv in bui.client.servers:
                out = {
                    "name": serv,
                    "stats": {"total": 0, "totsize": 0, "os": {}},
                    "number": 0,
                }
                if check and not mask.is_server_allowed(current_user, serv):
                    continue
                try:
                    clients = bui.client.get_all_clients(agent=serv, last_attempt=False)
                except BUIserverException:
                    continue
                if check:
                    clients = [
                        x
                        for x in clients
                        if mask.is_client_allowed(current_user, x["name"], serv)
                    ]

                j = bui.client.get_clients_report(clients, serv)
                if "clients" not in j or "backups" not in j:
                    continue

                os = {}
                for stats in j["clients"]:
                    for key in ["total", "totsize"]:
                        out["stats"][key] += stats["stats"][key]
                    if stats["stats"]["os"] in os:
                        os[stats["stats"]["os"]] += 1
                    else:
                        os[stats["stats"]["os"]] = 1

                for key, val in os.items():
                    out["stats"]["os"][key] = val

                for bkp in j["backups"]:
                    out["number"] += bkp["number"]

                backups.append({"name": serv, "number": out["number"]})
                servers.append({"name": serv, "number": out["stats"]})

        except BUIserverException as e:
            self.abort(500, str(e))

        r["backups"] = backups
        r["servers"] = servers

        return r



# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui
    :platform: Unix
    :synopsis: Burp-UI main module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import warnings

from .app import create_app

warnings.simplefilter("always", RuntimeWarning)


# backward compatibility
init = create_app



import click

from flask import Flask
from flask.cli import FlaskGroup


def create_app():
    app = Flask("burp-ui")
    return app


@click.group(cls=FlaskGroup, create_app=create_app)
def burp_ui():
    """Burp-UI CLI"""
    click.echo("COUCOU")



# -*- coding: utf8 -*-
"""
.. module:: burpui.forms
    :platform: Unix
    :synopsis: Burp-UI forms module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from .ext.i18n import LANGUAGES, get_locale

from flask_wtf import FlaskForm
from flask_babel import lazy_gettext as __
from wtforms import StringField, PasswordField, BooleanField, SelectField, validators


class LoginForm(FlaskForm):
    username = StringField(__("Username"), [validators.DataRequired()])
    password = PasswordField(__("Password"), [validators.DataRequired()])
    language = SelectField(
        __("Language"), choices=list(LANGUAGES.items()), default=get_locale
    )
    remember = BooleanField(__("Remember me"), [validators.Optional()])



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.limit
    :platform: Unix
    :synopsis: Burp-UI external Limit module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address

limiter = Limiter(key_func=get_remote_address)



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.backend.burp.constant
    :platform: Unix
    :synopsis: Burp-UI burp constant module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""

BURP_MINIMAL_VERSION = "burp-2.0.18"
BURP_LIST_BATCH = "2.0.48"
BURP_REVERSE_COUNTERS = "2.1.6"
BURP_STATUS_FORMAT_V2 = BURP_BIND_MULTIPLE = "2.1.10"
BURP_STATUS_DELIMITER = BURP_LISTEN_OPTION = "2.2.10"



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.backend.burp2
    :platform: Unix
    :synopsis: Burp-UI burp2 backend module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import re
import os
import time
import json

from collections import OrderedDict

from .burp1 import Burp as Burp1
from .interface import BUIbackend
from .utils.burp2 import Monitor
from .utils.constant import BURP_REVERSE_COUNTERS, BURP_STATUS_FORMAT_V2
from ..parser.burp2 import Parser
from ...utils import human_readable as _hr, utc_to_local
from ...exceptions import BUIserverException
from ..._compat import to_unicode


# Some functions are the same as in Burp1 backend
class Burp(Burp1):
    """The :class:`burpui.misc.backend.burp2.Burp` class provides a consistent
    backend for ``burp-2`` servers.

    It extends the :class:`burpui.misc.backend.burp1.Burp` class because a few
    functions can be reused. The rest is just overrided.

    :param server: ``Burp-UI`` server instance in order to access logger
                   and/or some global settings
    :type server: :class:`burpui.engines.server.BUIServer`

    :param conf: Configuration to use
    :type conf: :class:`burpui.config.BUIConfig`
    """

    # backend version
    _vers = 2
    # cache to store the guessed OS
    _os_cache = {}

    def __init__(self, server=None, conf=None):
        """
        :param server: ``Burp-UI`` server instance in order to access logger
                       and/or some global settings
        :type server: :class:`burpui.engines.server.BUIServer`

        :param conf: Configuration to use
        :type conf: :class:`burpui.config.BUIConfig`
        """

        BUIbackend.__init__(self, server, conf)

        self.monitor = Monitor(self.burpbin, self.burpconfcli, self.app, self.timeout)

        self.batch_list_supported = self.monitor.batch_list_supported

        self.parser = Parser(self)

        self.logger.info(f"burp binary: {self.burpbin}")
        self.logger.info(f"strip binary: {self.stripbin}")
        self.logger.info(f"burp conf cli: {self.burpconfcli}")
        self.logger.info(f"burp conf srv: {self.burpconfsrv}")
        self.logger.info(f"command timeout: {self.timeout}")
        self.logger.info(f"tmpdir: {self.tmpdir}")
        self.logger.info(f"zip64: {self.zip64}")
        self.logger.info(f"includes: {self.includes}")
        self.logger.info(f"enforce: {self.enforce}")
        self.logger.info(f"revoke: {self.revoke}")
        self.logger.info(f"client version: {self.client_version}")
        self.logger.info(f"server version: {self.server_version}")

    @property
    def client_version(self):
        return self.monitor.client_version

    @property
    def server_version(self):
        return self.monitor.server_version

    @staticmethod
    def _human_st_mode(mode):
        """Convert the st_mode returned by stat in human readable (ls-like)
        format
        """
        hur = ""
        if os.path.stat.S_ISREG(mode):
            hur = "-"
        elif os.path.stat.S_ISLNK(mode):
            hur = "l"
        elif os.path.stat.S_ISSOCK(mode):
            hur = "s"
        elif os.path.stat.S_ISDIR(mode):
            hur = "d"
        elif os.path.stat.S_ISBLK(mode):
            hur = "b"
        elif os.path.stat.S_ISFIFO(mode):
            hur = "p"
        elif os.path.stat.S_ISCHR(mode):
            hur = "c"
        else:
            hur = "-"

        for who in "USR", "GRP", "OTH":
            for perm in "R", "W", "X":
                if mode & getattr(os.path.stat, "S_I" + perm + who):
                    hur += perm.lower()
                else:
                    hur += "-"

        return hur

    def statistics(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.statistics`"""
        return {
            "alive": self.monitor.alive,
            "server_version": self.server_version,
            "client_version": self.client_version,
        }

    def status(self, query="c:\n", timeout=None, cache=True, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.status`"""
        return self.monitor.status(query, timeout, cache)

    def _get_backup_logs(self, number, client, forward=False, deep=False):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_backup_logs`
        """
        ret = {}
        ret2 = {}
        if not client or not number:
            return ret

        query = self.status("c:{0}:b:{1}\n".format(client, number))
        if not query:
            return ret
        try:
            logs = query["clients"][0]["backups"][0]["logs"]["list"]
        except KeyError:
            self.logger.warning("No logs found")
            return ret
        if "backup_stats" in logs:
            ret = self._parse_backup_stats(number, client, forward)
        if "backup" in logs and deep:
            ret2 = self._parse_backup_log(number, client)

        ret.update(ret2)

        ret["encrypted"] = False
        if "files_enc" in ret and ret["files_enc"]["total"] > 0:
            ret["encrypted"] = True
        return ret

    @staticmethod
    def _do_parse_backup_log(data, client):
        # tests ordered as the logs order
        ret = OrderedDict()
        ret["client_version"] = None
        ret["protocol"] = 1
        ret["is_windows"] = False
        ret["server_version"] = None
        if not data:
            return ret
        try:
            log = data["clients"][0]["backups"][0]["logs"]["backup"]
        except KeyError:
            # Assume protocol 1 in all cases unless explicitly found Protocol 2
            return ret
        # pre-compile regex since they'll be called on every log line
        regex = {
            "client_version": re.compile(r"Client version: (\d+\.\d+\.\d+)$"),
            "server_version": re.compile(
                r"WARNING: Client '{}' version '\d+\.\d+\.\d+' does not match server version '(\d+\.\d+\.\d+)'. An upgrade is recommended.$".format(
                    client
                )
            ),
            "protocol": re.compile(r"Protocol: (\d)$"),
            "is_windows": re.compile(r"Client is Windows$"),
        }
        expressions_list = list(ret.keys())
        catching_expressions = ["client_version", "server_version", "protocol"]
        casting_expressions = {
            "protocol": int,
        }

        def __dummy(val):
            return val

        for line in log:
            expressions = expressions_list
            for expression in expressions:
                if expression in catching_expressions:
                    catch = regex[expression].search(line)
                    if catch:
                        cast = casting_expressions.get(expression, __dummy)
                        ret[expression] = cast(catch.group(1))
                        # don't search this expression twice
                        expressions_list.remove(expression)
                        break
                else:
                    if expression in regex and regex[expression].search(line):
                        ret[expression] = True
                        # don't search this expression twice
                        expressions_list.remove(expression)
                        break
        return ret

    def _parse_backup_log(self, number, client):
        """The :func:`burpui.misc.backend.burp2.Burp._parse_backup_log`
        function helps you determine if the backup is protocol 2 or 1 and various
        useful details.

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :returns: a dict with some useful details
        """
        data = self.status("c:{0}:b:{1}:l:backup\n".format(client, number))
        return self._do_parse_backup_log(data, client)

    def _do_parse_backup_stats(
        self, data, result, number, client, forward=False, agent=None
    ):
        ret = {}
        translate = {
            "time_start": "start",
            "time_end": "end",
            "time_taken": "duration",
            "bytes": "totsize",
            "bytes_received": "received",
            "bytes_estimated": "estimated_bytes",
            "files": "files",
            "files_encrypted": "files_enc",
            "directories": "dir",
            "soft_links": "softlink",
            "hard_links": "hardlink",
            "meta_data": "meta",
            "meta_data_encrypted": "meta_enc",
            "special_files": "special",
            "efs_files": "efs",
            "vss_headers": "vssheader",
            "vss_headers_encrypted": "vssheader_enc",
            "vss_footers": "vssfooter",
            "vss_footers_encrypted": "vssfooter_enc",
            "total": "total",
            "grand_total": "total",
        }
        counts = {
            "new": "count",
            "changed": "changed",
            "unchanged": "same",
            "deleted": "deleted",
            "total": "scanned",
            "scanned": "scanned",
        }
        single = [
            "time_start",
            "time_end",
            "time_taken",
            "bytes_received",
            "bytes_estimated",
            "bytes",
        ]
        if not data:
            return ret
        try:
            back = data["clients"][0]["backups"][0]
        except KeyError:
            self.logger.warning("No backup found")
            return ret
        if "backup_stats" not in back["logs"]:
            self.logger.warning("No stats found for backup")
            return ret
        stats = None
        try:
            stats = json.loads("".join(back["logs"]["backup_stats"]))
        except:
            stats = back["logs"]["backup_stats"]
        if not stats:
            return ret
        # server was upgraded but backup comes from an older version
        if "counters" not in stats:
            return super(Burp, self)._parse_backup_stats(
                number, client, forward, stats, agent
            )
        counters = stats["counters"]
        for counter in counters:
            name = counter["name"]
            if name in translate:
                name = translate[name]
            if counter["name"] in single:
                result[name] = counter["count"]
            else:
                result[name] = {}
                for (key, val) in counts.items():
                    if val in counter:
                        result[name][key] = counter[val]
                    else:
                        result[name][key] = 0
        if "start" in result and "end" in result:
            result["duration"] = result["end"] - result["start"]
            # convert utc timestamp to local
            # example: 1468850307 -> 1468857507
            result["start"] = utc_to_local(result["start"])
            result["end"] = utc_to_local(result["end"])

        # Needed for graphs
        if "received" not in result:
            result["received"] = 1

        return result

    def _parse_backup_stats(self, number, client, forward=False, agent=None):
        """The :func:`burpui.misc.backend.burp2.Burp._parse_backup_stats`
        function is used to parse the burp logs.

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :param forward: Is the client name needed in later process
        :type forward: bool

        :param agent: What server to ask (only in multi-agent mode)
        :type agent: str

        :returns: Dict containing the backup log
        """
        backup = {"os": self._guess_os(client), "number": int(number)}
        if forward:
            backup["name"] = client
        query = self.status("c:{0}:b:{1}:l:backup_stats\n".format(client, number))
        return self._do_parse_backup_stats(
            query, backup, number, client, forward, agent
        )

    # TODO: support old clients
    # NOTE: this should now be partly done since we fallback to the Burp1 code
    # def _parse_backup_log(self, fh, number, client=None, agent=None):
    #    """
    #    parse_backup_log parses the log.gz of a given backup and returns a
    #    dict containing different stats used to render the charts in the
    #    reporting view
    #    """
    #    return {}

    # def get_clients_report(self, clients, agent=None):

    def get_counters(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_counters`"""
        ret = {}
        query = self.status("c:{0}\n".format(name), cache=False)
        # check the status returned something
        if not query:
            return ret

        try:
            client = query["clients"][0]
        except KeyError:
            self.logger.warning("Client not found")
            return ret
        return self._do_get_counters(client)

    def _do_get_counters(self, data):
        ret = {}
        client = data
        # check the client is currently backing-up
        if "run_status" not in client or client["run_status"] != "running":
            return ret

        backup = None
        phases = ["working", "finishing"]
        try:
            for child in client["children"]:
                if "action" in child and child["action"] == "backup":
                    backup = child
                    break
        except KeyError:
            for back in client["backups"]:
                if "flags" in back and any(x in back["flags"] for x in phases):
                    backup = back
                    break
        # check we found a working backup
        if not backup:
            return ret

        # list of single counters (type CNTR_SINGLE_FIELD in cntr.c)
        single = [
            "bytes_estimated",
            "bytes",
            "bytes_received",
            "bytes_sent",
            "time_start",
            "time_end",
            "warnings",
            "errors",
        ]

        # translation table to be compatible with burp1
        def translate(cntr):
            translate_table = {"bytes_estimated": "estimated_bytes"}
            try:
                return translate_table[cntr]
            except KeyError:
                return cntr

        for counter in backup.get("counters", {}):
            name = translate(counter["name"])
            if counter["name"] not in single:
                # Prior burp-2.1.6 some counters are reversed
                # See https://github.com/grke/burp/commit/adeb3ad68477303991a393fa7cd36bc94ff6b429
                if self.server_version and self.server_version < BURP_REVERSE_COUNTERS:
                    ret[name] = [
                        counter["count"],
                        counter["same"],  # reversed
                        counter["changed"],  # reversed
                        counter["deleted"],
                        counter["scanned"],
                    ]
                else:
                    ret[name] = [
                        counter["count"],
                        counter["changed"],
                        counter["same"],
                        counter["deleted"],
                        counter["scanned"],
                    ]
            else:
                ret[name] = counter["count"]

        if "phase" in backup:
            ret["phase"] = backup["phase"]
        else:
            for phase in phases:
                if phase in backup.get("flags", []):
                    ret["phase"] = phase
                    break

        if "bytes" not in ret:
            ret["bytes"] = 0
        if set(["time_start", "estimated_bytes", "bytes"]) <= set(ret.keys()):
            try:
                diff = time.time() - int(ret["time_start"])
                byteswant = int(ret["estimated_bytes"])
                bytesgot = int(ret["bytes"])
                bytespersec = bytesgot / diff
                bytesleft = byteswant - bytesgot
                ret["speed"] = bytespersec
                if bytespersec > 0:
                    timeleft = int(bytesleft / bytespersec)
                    ret["timeleft"] = timeleft
                else:
                    ret["timeleft"] = -1
            except:
                ret["timeleft"] = -1
        try:
            ret["percent"] = round(
                float(ret["bytes"]) / float(ret["estimated_bytes"]) * 100
            )
        except (ZeroDivisionError, KeyError):
            # You know... division by 0
            ret["percent"] = 0

        return ret

    def is_backup_running(self, name=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_backup_running`
        """
        if not name:
            return False
        try:
            query = self.status("c:{0}\n".format(name))
        except BUIserverException:
            return False
        return self._do_is_backup_running(query)

    def _do_is_backup_running(self, data):
        if data:
            try:
                return data["clients"][0]["run_status"] in ["running"]
            except KeyError:
                pass
        return False

    def is_one_backup_running(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_one_backup_running`
        """
        ret = []
        try:
            clients = self.get_all_clients(last_attempt=False)
        except BUIserverException:
            return ret
        return self._do_is_one_backup_running(clients)

    def _do_is_one_backup_running(self, data):
        ret = []
        for client in data:
            if client["state"] in ["running"]:
                ret.append(client["name"])
        return ret

    def _status_human_readable(self, status):
        """The label has changed in burp2, we override it to be compatible with
        burp1's format

        :param status: The status returned by the burp2 server
        :type status: str

        :returns: burp1 status compatible
        """
        if not status:
            return None
        if status == "c crashed":
            return "client crashed"
        if status == "s crashed":
            return "server crashed"
        return status

    def _get_last_backup(self, name, working=True):
        """Return the last backup of a given client

        :param name: Name of the client
        :type name: str

        :param working: Also return uncomplete backups
        :type working: bool

        :returns: The last backup
        """
        try:
            clients = self.status("c:{}".format(name))
            client = clients["clients"][0]
            i = 0
            while True:
                ret = client["backups"][i]
                if not working and "working" in ret["flags"]:
                    i += 1
                    continue
                return ret
        except (KeyError, TypeError, IndexError, BUIserverException):
            return None

    def _guess_os(self, name):
        """Return the OS of the given client based on the magic *os* label

        :param name: Name of the client
        :type name: str

        :returns: The guessed OS of the client

        ::

            grep label /etc/burp/clientconfdir/toto
            label = os: Darwin OS
        """
        ret = "Unknown"
        if name in self._os_cache:
            return self._os_cache[name]

        labels = self.get_client_labels(name)
        OSES = []

        for label in labels:
            if re.match("os:", label, re.IGNORECASE):
                _os = label.split(":", 1)[1].strip()
                if _os not in OSES:
                    OSES.append(_os)

        if OSES:
            ret = OSES[-1]
        else:
            # more aggressive check
            last = self._get_last_backup(name, False)
            if last:
                try:
                    tree = self.get_tree(name, last["number"])

                    if tree[0]["name"] != "/":
                        ret = "Windows"
                    else:
                        ret = "Unix/Linux"
                except (IndexError, KeyError, BUIserverException):
                    pass

        self._os_cache[name] = ret
        return ret

    def get_all_clients(self, agent=None, last_attempt=True):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_all_clients`
        """
        ret = []
        query = self.status()
        if not query or "clients" not in query:
            return ret
        clients = query["clients"]
        for client in clients:
            cli = {}
            cli["name"] = client["name"]
            cli["state"] = self._status_human_readable(client["run_status"])
            infos = client["backups"]
            if cli["state"] in ["running"]:
                cli["last"] = "now"
                cli["last_attempt"] = "now"
            elif not infos:
                cli["last"] = "never"
                cli["last_attempt"] = "never"
            else:
                convert = True
                infos = infos[0]
                if self.server_version and self.server_version < BURP_STATUS_FORMAT_V2:
                    cli["last"] = infos["timestamp"]
                    convert = False
                # only do deep inspection when server >= BURP_STATUS_FORMAT_V2
                elif self.deep_inspection:
                    logs = self.get_backup_logs(infos["number"], client["name"])
                    cli["last"] = logs["start"]
                else:
                    cli["last"] = utc_to_local(infos["timestamp"])
                if last_attempt:
                    last_backup = self._get_last_backup(client["name"])
                    if last_attempt:
                        if convert:
                            cli["last_attempt"] = utc_to_local(last_backup["timestamp"])
                        else:
                            cli["last_attempt"] = last_backup["timestamp"]

            ret.append(cli)
        return ret

    def get_client_status(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_status`"""
        ret = {}
        if not name:
            return ret
        query = self.status("c:{0}\n".format(name))
        if not query:
            return ret
        try:
            client = query["clients"][0]
        except (KeyError, IndexError):
            self.logger.warning("Client not found")
            return ret
        return self._do_get_client_status(client)

    def _do_get_client_status(self, data):
        ret = {}
        client = data
        ret["state"] = self._status_human_readable(client["run_status"])
        infos = client["backups"]
        if ret["state"] in ["running"]:
            try:
                ret["phase"] = client["phase"]
            except KeyError:
                for child in client.get("children", []):
                    if "action" in child and child["action"] == "backup":
                        ret["phase"] = child["phase"]
                        break
            counters = self._do_get_counters(client)
            if "percent" in counters:
                ret["percent"] = counters["percent"]
            else:
                ret["percent"] = 0
            ret["last"] = "now"
        elif not infos:
            ret["last"] = "never"
        else:
            infos = infos[0]
            ret["last"] = infos["timestamp"]
        return ret

    def get_client(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client`"""
        return self.get_client_filtered(name)

    def get_client_filtered(
        self, name=None, limit=-1, page=None, start=None, end=None, agent=None
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_filtered`"""
        ret = []
        if not name:
            return ret
        query = self.status("c:{0}\n".format(name))
        if not query:
            return ret
        try:
            backups = query["clients"][0]["backups"]
        except (KeyError, IndexError):
            self.logger.warning("Client not found")
            return ret
        for idx, backup in enumerate(backups):
            # skip the first elements if we are in a page
            if page and page > 1 and limit > 0:
                if idx < (page - 1) * limit:
                    continue
            back = {}
            # skip running backups since data will be inconsistent
            if "flags" in backup and "working" in backup["flags"]:
                continue
            back["number"] = backup["number"]
            if "flags" in backup and "deletable" in backup["flags"]:
                back["deletable"] = True
            else:
                back["deletable"] = False
            back["date"] = backup["timestamp"]
            # skip backups before "start"
            if start and backup["timestamp"] < start:
                continue
            # skip backups after "end"
            if end and backup["timestamp"] > end:
                continue

            def __get_log(client, bkp, res):
                log = self.get_backup_logs(bkp["number"], client)
                try:
                    res["encrypted"] = log["encrypted"]
                    try:
                        res["received"] = log["received"]
                    except KeyError:
                        res["received"] = 0
                    try:
                        res["size"] = log["totsize"]
                    except KeyError:
                        res["size"] = 0
                    res["end"] = log["end"]
                    # override date since the timestamp is odd
                    res["date"] = log["start"]
                except Exception:
                    self.logger.warning("Unable to parse logs")
                    return None
                return res

            with_log = __get_log(name, backup, back)
            if with_log:
                ret.append(with_log)

            # stop after "limit" elements
            if page and page > 1 and limit > 0:
                if idx >= page * limit:
                    break
            elif limit > 0 and idx >= limit:
                break

        # Here we need to reverse the array so the backups are sorted by num
        # ASC
        ret.reverse()
        return ret

    def is_backup_deletable(self, name=None, backup=None, agent=None):
        """Check if a given backup is deletable"""
        if not name or not backup:
            return False
        query = self.status("c:{0}:b:{1}\n".format(name, backup))
        if not query:
            return False
        return self._do_is_backup_deletable(query)

    def _do_is_backup_deletable(self, data):
        query = data
        try:
            flags = query["clients"][0]["backups"][0]["flags"]
            return "deletable" in flags
        except KeyError:
            return False

    def _format_tree(self, data, top, level):
        ret = []
        if not data:
            return ret
        try:
            backup = data["clients"][0]["backups"][0]
        except KeyError:
            return ret
        for entry in backup["browse"]["entries"]:
            data = {}
            base = None
            dirn = None
            if top == "*":
                base = os.path.basename(entry["name"])
                dirn = os.path.dirname(entry["name"])
            if entry["name"] == ".":
                continue
            else:
                data["name"] = base or entry["name"]
            data["mode"] = self._human_st_mode(entry["mode"])
            if re.match("^(d|l)", data["mode"]):
                data["type"] = "d"
                data["folder"] = True
            else:
                data["type"] = "f"
                data["folder"] = False
            data["inodes"] = entry["nlink"]
            data["uid"] = entry["uid"]
            data["gid"] = entry["gid"]
            data["parent"] = dirn or top
            data["size"] = "{0:.1eM}".format(_hr(entry["size"]))
            data["date"] = entry["mtime"]
            data["fullname"] = (
                os.path.join(top, entry["name"]) if top != "*" else entry["name"]
            )
            data["level"] = level
            data["children"] = []
            ret.append(data)
        return ret

    def get_tree(self, name=None, backup=None, root=None, level=-1, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_tree`"""
        if not name or not backup:
            return []
        if not root:
            top = ""
        else:
            top = to_unicode(root)

        # we know this operation may take a while so we arbitrary increase the
        # read timeout
        timeout = None
        if top == "*":
            timeout = max(self.timeout, 300)

        query = self.status("c:{0}:b:{1}:p:{2}\n".format(name, backup, top), timeout)
        return self._format_tree(query, top, level)

    def get_client_version(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_client_version`
        """
        return self.client_version

    def get_server_version(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_server_version`
        """
        if not self.server_version:
            self.status()
        return self.server_version

    def get_client_labels(self, client=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_client_labels`
        """
        ret = []
        if not client:
            return ret
        # micro optimization since the status results are cached in memory for a
        # couple seconds, using the same global query and iterating over it
        # will be more efficient than filtering burp-side
        query = self.status("c:\n")
        if not query:
            return ret
        try:
            for cli in query["clients"]:
                if cli["name"] == client:
                    return cli["labels"]
        except KeyError:
            return ret

    # Same as in Burp1 backend
    # def restore_files(
    #     self,
    #     name=None,
    #     backup=None,
    #     files=None,
    #     strip=None,
    #     archive='zip',
    #     password=None,
    #     agent=None):

    # def read_conf_cli(self, agent=None):

    # def read_conf_srv(self, agent=None):

    # def store_conf_cli(self, data, agent=None):

    # def store_conf_srv(self, data, agent=None):

    # def get_parser_attr(self, attr=None, agent=None):



# -*- coding: utf8 -*-
"""
.. module:: burpui._compat
    :platform: Unix
    :synopsis: Burp-UI compatibility module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import pickle  # noqa

from urllib.parse import unquote, quote, urlparse, urljoin  # noqa

text_type = str
string_types = (str,)


def to_bytes(text):
    """Transform string to bytes."""
    if isinstance(text, text_type):
        text = text.encode("utf-8")
    return text or b""


def to_unicode(input_bytes, encoding="utf-8"):
    """Decodes input_bytes to text if needed."""
    if not isinstance(input_bytes, string_types):
        input_bytes = input_bytes.decode(encoding)
    return input_bytes or ""


# maps module name -> attribute name -> original item
# e.g. "time" -> "sleep" -> built-in function sleep
saved = {}


# Borrowed from gevent in order to patch json
def patch_item(module, attr, newitem, newmodule=None):
    NONE = object()
    olditem = getattr(module, attr, NONE)
    if olditem is not NONE:
        saved.setdefault(module.__name__, {}).setdefault(attr, olditem)
        if newmodule and not getattr(newmodule, "ori_" + attr, None):
            setattr(newmodule, "ori_" + attr, olditem)
    if not getattr(newmodule, "ori_" + attr, None):
        setattr(module, attr, newitem)


def patch_module(name, items=None):
    toimport = items or []
    mod = __name__
    if "." in mod:
        mod = mod.split(".")[0]
    replace_module = __import__("{}._{}".format(mod, name), fromlist=toimport)
    module_name = name
    module = __import__(module_name)
    if items is None:
        items = getattr(replace_module, "__implements__", None)
        if items is None:
            raise AttributeError("%r does not have __implements__" % replace_module)
    for attr in items:
        patch_item(module, attr, getattr(replace_module, attr), replace_module)


def patch_json():
    try:
        import ujson  # noqa
    except ImportError:
        # ujson is not available, we won't patch anything
        return
    patch_module("json", ["dumps", "loads"])



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.audit.interface
    :platform: Unix
    :synopsis: Burp-UI audit interface.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import logging

from abc import ABCMeta, abstractmethod


class BUIauditLogger(object, metaclass=ABCMeta):
    """The :class:`burpui.misc.audit.interface.BUIauditLogger` class defines the audit
    Logger interface.
    """

    _level = -1

    @property
    def level(self):
        return self._level

    def debug(self, message, *args, **kwargs):
        if logging.DEBUG >= self.level:
            self.log(logging.DEBUG, message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        if logging.INFO >= self.level:
            self.log(logging.INFO, message, *args, **kwargs)

    def warn(self, message, *args, **kwargs):
        if logging.WARN >= self.level:
            self.log(logging.WARN, message, *args, **kwargs)

    warning = warn

    def err(self, message, *args, **kwargs):
        if logging.ERROR >= self.level:
            self.log(logging.ERROR, message, *args, **kwargs)

    error = err

    def critical(self, message, *args, **kwargs):
        if logging.CRITICAL >= self.level:
            self.log(logging.CRITICAL, message, *args, **kwargs)

    @abstractmethod
    def log(self, level, message, *args, **kwargs):
        pass


class BUIaudit(object, metaclass=ABCMeta):
    """The :class:`burpui.misc.audit.interface.BUIaudit` class defines the audit
    interface.

    :param app: Instance of the app we are running in
    :type app: :class:`burpui.engines.server.BUIServer`
    """

    priority = 0

    name = None
    _logger = None

    def __init__(self, app):
        self.app = app

    @property
    def logger(self) -> BUIauditLogger:
        """:rtype: class:`BUIauditLogger`"""
        return self._logger
